# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "IA Model Format (.ia)",
    "author": "dom107",
    "version": (1, 2, 6),
    "blender": (2, 63, 0),
    "location": "File > Export > TS20xx IA (.ia)",
    "description": "Export IA Model Format (.ia) for Railworks",
    "warning": "",
    "tracker_url": "",
    "category": "Import-Export"}

EXPORTER_VERSION = 'IA exporter version 1.2.6' # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    
######################################################
# Data Structures
######################################################

# array size predefinitions
IAF_MAX_ANIMTRACK_NAME       = 32
IAF_MAX_HEADER_COMMENT       = 64
IAF_MAX_TRIGGER_DATA_NAME    = 32

IA_HEADER_SIZE = 104
IA_ANIMATIONNODE_SIZE = 44
IA_MOTIONCONTROLLER_SIZE = 12
IA_MOTIONSET_SIZE = 16

MAXFRAMERATEMULTPLIER = 10

import bpy
import struct
import os

from bpy.props import StringProperty, EnumProperty, BoolProperty

from mathutils import Vector, Matrix
import configparser
import sys
import time

# Container for the exporter settings
class IAExporterSettings:
    def __init__(self,
                 context,
                 FilePath,
                 use_selection=True,
                 visible_children=False,
                 Verbose=False):
        self.context = context
        self.FilePath = FilePath
        self.use_selection = use_selection
        self.visible_children = visible_children
        self.Verbose = Verbose
        
#-----------------------------------------------------------------------------
# CLASS _IAMotionController
#
# defines a set of keys creating motion over time - this may be for rotation,
# position, or whatever. 
class _IAMotionController:
    __slots__ = "mControllerType", "mKeySize", "mKeyStream"

    def __init__(self, mControllerType, mKeySize, mKeyStream):
        self.mControllerType = mControllerType # Type of data to expect in the key stream
        # 0: MotionVector4Frame / 1: MotionQuaternionFrame
        # List of keys, whatever they may be (defined by mControllerType)
        self.mKeySize = mKeySize # Size, in bytes, per key item (eg, 16 for svector4)
        self.mKeyStream = mKeyStream

    def write(self, file):
        file.write(struct.pack('<2I', self.mControllerType, self.mKeySize))
        file.write(struct.pack('<I', self.mKeyStream))

    def dump_log(self, motion_controller_num, MotionControllerCount, offset):
        print('--------------------------------------------------------------------------------')
        print('IAfMotionControllers [ {:d}  /  {:d}  ] ({:d}) :'.format(motion_controller_num, MotionControllerCount, offset))
        print('------------------------------.')
        print("MotionControllerType          : {:d}".format(self.mControllerType))
        print("KeySize                       : {:d}".format(self.mKeySize))
        print("KeyStreamObjectStartOffset    : {:d}".format(self.mKeyStream))

    def dump_log2(self):
        print('------------------------------.')
        print("MotionControllerType          : {:d}".format(self.mControllerType))
        print("KeySize                       : {:d}".format(self.mKeySize))
        print("KeyStreamObjectStartOffset    : {:d}".format(self.mKeyStream))


#-----------------------------------------------------------------------------
# CLASS _IAAnimationNode
#
# defines grouped controllers of animation for a single node in a shape 
# hierarchy - each node will have a corresponding _IAAnimationNode (with matching name)
# that defines the animation result over the total animation time
class _IAAnimationNode:
    __slots__ = "mName", "mNumControllers", "mMotionControllers", "mFlags"

    def __init__(self, mName, mNumControllers, mMotionControllerCount, mMotionControllers):
        # name of track, will be matched to node to be animated by this data
        self.mName = str.encode(mName)
        # number of controllers present (2, usually, for pos + rot)
        self.mNumControllers = mNumControllers
        # list of controllers in use
        self.mMotionControllers = mMotionControllers
        # generic flag field
        self.mFlags = 0
        
    def get_size(self):
        return IA_ANIMATIONNODE_SIZE

    def write(self, file):
        block_format = '<{:d}s2I'.format(IAF_MAX_ANIMTRACK_NAME)
        file.write(struct.pack(block_format, self.mName, self.mNumControllers, self.mMotionControllers))
        file.write(struct.pack('<I', self.mFlags))

    def dump_log(self, anim_node_num, AnimationNodeCount, offset, verbose):
        if verbose:
            print('--------------------------------------------------------------------------------')
            print('IAfAnimationNodes [ {:d}  /  {:d}  ] ({:d}) :'.format(anim_node_num, AnimationNodeCount, offset))
            print('------------------------------.')
            print("Name                          : {:s}".format(self.mName.decode()))
            print("MotionControllerCount         : {:d}".format(self.mNumControllers))
            print("MotionControllerListStartOffset : {:d}".format(self.mMotionControllers))
            print("Flags                         : {:d}".format(self.mFlags))
        else:
            print("  {:s}".format(self.mName.decode()))

    def dump_log2(self):
        print('------------------------------.')
        print("Name                          : {:s}".format(self.mName.decode()))
        print("MotionControllerCount         : {:d}".format(self.mNumControllers))
        print("MotionControllerListStartOffset : {:d}".format(self.mMotionControllers))
        print("Flags                         : {:d}".format(self.mFlags))


#-----------------------------------------------------------------------------
# CLASS _IATrigger
#
# defines an trigger event at a given time (in seconds) e.g. audio playback.
class _IATrigger:
    __slots__ = "mFrameTime", "mTriggerData"

    def __init__(self, mFrameTime, mTriggerData):
        # Time (in seconds) that this trigger is locked to
        self.mFrameTime = mFrameTime
        # Trigger data.
        # This relates to externally defined events such as descriptor names,
        # sound effect names special effect names etc.
        self.mTriggerData = mTriggerData

    def write(self, file):
        block_format = '<I{:d}s'.format(IAF_MAX_TRIGGER_DATA_NAME)
        file.write(struct.pack(block_format, self.mFrameTime, self.mTriggerData))

    def dump_log(self):
        print("  X                             : {:d}".format(self.mFrameTime))
        print("  X                             : {:s}".format(self.mTriggerData))


#-----------------------------------------------------------------------------
# CLASS MotionVector4Frame
#
class MotionVector4Frame:
    __slots__ = "x", "y", "z", "w"

    def __init__(self, x, y, z, w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def get_size(self):
        return IA_MOTIONSET_SIZE

    def write(self, file):
        file.write(struct.pack('<4f', self.x, self.y, self.z, self.w))

    def dump_log(self, frame_num):
        print("MotionVector4Frame{:03d} ----------.".format(frame_num))
        print("  X                             : {:f}".format(self.x))
        print("  Y                             : {:f}".format(self.y))
        print("  Z                             : {:f}".format(self.z))
        print("  Weight                        : {:f}".format(self.w))       

#-----------------------------------------------------------------------------
# CLASS MotionQuaternionFrame
#
class MotionQuaternionFrame:
    __slots__ = "a", "b", "c", "d"

    def __init__(self, a, b, c, d):
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    def get_size(self):
        return IA_MOTIONSET_SIZE

    def write(self, file):
        file.write(struct.pack('<4f', self.a, self.b, self.c, self.d))

    def dump_log(self, frame_num):
        print("MotionQuaternionFrame{:03d} -------.".format(frame_num))
        print("  a                             : {:f}".format(self.a))
        print("  b                             : {:f}".format(self.b))
        print("  c                             : {:f}".format(self.c))
        print("  d                             : {:f}".format(self.d))      

#-----------------------------------------------------------------------------
# CLASS _IATriggerTimeline
#
# Defines a global trigger timeline, used to tag when in the animation
# that triggers should be fired.
# Triggers include sounds, special effects etc.
class _IATriggerTimeline:
    __slots__ = "mNumKeys", "mTriggerList"

    def __init__(self, mNumKeys, mTriggerList):
        # number of triggers present
        self.mNumKeys = mNumKeys
        # list of triggers, mNumKeys long
        self.mTriggerList = mTriggerList

    def write(self, file):
        file.write(struct.pack('<I', self.mNumKeys))
        file.write(struct.pack('<I', self.mTriggerList))

    def dump_log(self):
        print("TriggerTimeline ----------------.")
        print("  TriggerKeyCount               : {:d}".format(self.mNumKeys))
        print("  TriggerListStartOffset        : {:d}".format(self.mTriggerList))


#-----------------------------------------------------------------------------
# CLASS _IAFHeader
#
# main file header, always written first in the IA file stream
class _IAFHeader(object):
# Main file header, always written first in the IG file stream
    __slots__ = "mMagic", "mVersion", "mCommentField", "mDataStart", "mTotalFrames", "mSampleRate", "mTriggersTimeline", "mNumAnimationNodes", "mAnimationNodes", "mDataEnd"

    def __init__(self, mCommentField, mDataStart, mTotalFrames, mSampleRate, mTriggersTimeline, mNumAnimationNodes, mAnimationNodes, mDataEnd):
        self.mMagic = b'KIAF'
        self.mVersion = 100 # current version of IA file
        # generic comment / description field
        self.mCommentField = str.encode(mCommentField)
        self.mDataStart = mDataStart # tags beginning of file data
        # animation setup
        self.mTotalFrames = mTotalFrames # total frames in the whole animation
        self.mSampleRate = mSampleRate # sample rate in Hz used to get data from controllers (default 300)
        # Stream of trigger data (e.g. audio triggers).
        self.mTriggersTimeline = mTriggersTimeline # class _IATriggerTimeline
        # animation tracks for all shapes represented by this IA file
        self.mNumAnimationNodes = mNumAnimationNodes # eg, num shapes dumped
        self.mAnimationNodes = mAnimationNodes
        self.mDataEnd = mDataEnd

    def write(self, file):
        file.write(struct.pack('<4sI', self.mMagic, self.mVersion))
        block_format = "<{:d}sI".format(IAF_MAX_HEADER_COMMENT)
        file.write(struct.pack(block_format, self.mCommentField, self.mDataStart))
        file.write(struct.pack('2I', self.mTotalFrames, self.mSampleRate))
        self.mTriggersTimeline.write(file)
        file.write(struct.pack('I', self.mNumAnimationNodes))
        file.write(struct.pack('I', self.mAnimationNodes))
        file.write(struct.pack('I', self.mDataEnd))

    def dump_log(self):
        print("================================================================================")
        print("IAfHeaders                    : [ 1 ]")
        print("--------------------------------------------------------------------------------")
        print("IAfHeaders [ 1  /  1  ] (0)  ")
        print("------------------------------.")
        print("MagicId                       : '{:s}'".format(self.mMagic.decode()))
        print("Version                       : {:d}".format(self.mVersion))
        print("CommentField                  : '{:s}'".format(self.mCommentField.decode()))
        print("DataStart                     : {:d}".format(self.mDataStart)) # A changer        
        print("TotalFrames                   : {:d}".format(self.mTotalFrames))
        print("SampleRate                    : {:d}".format(self.mSampleRate))
        self.mTriggersTimeline.dump_log()
        print("AnimationNodeCount            : {:d}".format(self.mNumAnimationNodes))
        print("AnimationNodeListStartOffset  : {:d}".format(self.mAnimationNodes))
        print("DataEnd                       : {:d}".format(self.mDataEnd))
        
                
def print_config(config, config_name):
    print('\n Configuration file: ', config_name)
    for section in config.sections():
        print('------ ',section,' ------')
        for option in config.items(section):
            print(option)
    print('\n')                

def print_config_section(config, config_name, section_name):
    print('\n Extract from configuration file: ', config_name)
    for section in config.sections():
        if section == section_name:
            print('------ ',section,' ------')
            for option in config.items(section):
                print(option)
    print('\n')


def rename_object(Name):
# Rename to be compliant to 1_0100_name
    LODlevel = '1'
    LODdist = '1000'
    name_split = Name.split('_')
    if len(name_split) >= 2:
        if name_split[0].isdigit() and name_split[1].isdigit():
            LODlevel = name_split[0]
            LODdist = name_split[1]
            return Name
        else:
            return (LODlevel + '_' + LODdist + '_' + Name)
    else:
        return (LODlevel + '_' + LODdist + '_' + Name)


class object_wrapper(object):
    '''Class to gather animation ia objects for a Blender object'''

    __slots__ = "name", "animation_node", "motion_controller", "motion_vector", "motion_quaternion"

    def __init__(self, name, animation_node, motion_controller, motion_vector, motion_quaternion):
        self.name = name
        self.animation_node = animation_node # 1 automation node
        self.motion_controller = motion_controller # 2 motion controllers
        self.motion_vector = motion_vector
        self.motion_quaternion = motion_quaternion

    def dump(self):
        print("--- object_wrapper --- {:s}".format(self.name))
        self.animation_node.dump_log2()
        for mc in self.motion_controller:
            mc.dump_log2()


def restore_std_out(outfile, saveout):
    # restore stdout
    outfile.flush()
    outfile.flush()
    outfile.close()
    sys.stdout = saveout

def has_animation_in_hierachy(Object):
# Has animation or has animation as parent
    if (Object.animation_data or Object.constraints):
        return True
    if (Object.parent is not None):
        if (Object.parent.type == 'ARMATURE'):
            return True
    for obj_children in Object.children:
        if (obj_children.animation_data or obj_children.constraints):
            return True
    return False

    
def create_ia_objects(bpy, Object, parent, ia_objects_list, KeyframeCount, processed_objects_list, \
                        animation_node_index, motion_controller_index, FrameRateMultiplier, applied_translation, visible_children, scene):
    #print(" start create_ia_objects Object={:s} Object.type={:s}".format(Object.name, Object.type))
    if Object.type in {'MESH', 'EMPTY'}:
        if Object in processed_objects_list:
            #print(" already processed {:s} {:s}".format(Object.name, Object.type))
            return animation_node_index, motion_controller_index
            
        if parent is not None:
            applied_translation = Vector((0.0, 0.0, 0.0)) # It's a child. Don't move it as it's done in its parent data.

        wrapper_name = rename_object(Object.name) # Rename, if needed, to be compliant to x_nnn_name
        vector_frame_list = []
        quaternion_frame_list = []
        motion_controller_list= []
        current_frame = bpy.context.scene.frame_current
        #print(" KeyframeCount {:d}".format(KeyframeCount))
        for Frame in range(KeyframeCount):
            bpy.context.scene.frame_set(Frame + bpy.context.scene.frame_start)
            if parent is not None:
                #print(" ---- >>>> Object / Parent ", Object.name, parent.name)
                MyMatrix = parent.matrix_world.inverted() * Object.matrix_world
                #MyMatrix = Object.matrix_local
                Position = MyMatrix.to_translation() - applied_translation
                #print(" ---- >>>> Position ", Position)
            else:
                Position = Object.matrix_world.to_translation() - applied_translation
            new_vector = MotionVector4Frame(Position[0], Position[2], Position[1], 1.0)

            if parent is not None:
                MyMatrix = parent.matrix_world.inverted() * Object.matrix_world
                MyMatrix.translation = MyMatrix.to_translation() - applied_translation
            else:
                MyMatrix = Object.matrix_world.copy()
                MyMatrix.translation = MyMatrix.to_translation() - applied_translation
                #MyMatrix = MyMatrix.to_3x3()
            Rotation = MyMatrix.to_quaternion()
            new_quaternion = MotionQuaternionFrame(Rotation[1], Rotation[3], Rotation[2], Rotation[0])

            if Frame is not 0:
            # Add intermediary frames for smoothness
                for i in range(1, FrameRateMultiplier):
                    inter_vector = MotionVector4Frame(prev_position[0] + (Position[0]-prev_position[0])*i/FrameRateMultiplier, \
                                                        prev_position[2] + (Position[2]-prev_position[2])*i/FrameRateMultiplier, \
                                                        prev_position[1] + (Position[1]-prev_position[1])*i/FrameRateMultiplier, \
                                                        1.0)
                    inter_quaternion = MotionQuaternionFrame(prev_rotation[1] + (Rotation[1]-prev_rotation[1])*i/FrameRateMultiplier, \
                                                                prev_rotation[3] + (Rotation[3]-prev_rotation[3])*i/FrameRateMultiplier, \
                                                                prev_rotation[2] + (Rotation[2]-prev_rotation[2])*i/FrameRateMultiplier, \
                                                                prev_rotation[0] + (Rotation[0]-prev_rotation[0])*i/FrameRateMultiplier)
                    vector_frame_list.append(inter_vector)
                    quaternion_frame_list.append(inter_quaternion)

            vector_frame_list.append(new_vector)
            prev_position = Position
            quaternion_frame_list.append(new_quaternion)
            prev_rotation = Rotation
            
        # mControllerType, mKeySize, mKeyStream
        motion_controller_list.append(_IAMotionController(0, 16, motion_controller_index))
        motion_controller_list.append(_IAMotionController(1, 16, motion_controller_index + 1))
        
        animation_node = _IAAnimationNode(wrapper_name, 2, animation_node_index, motion_controller_index)

        animation_node_index += 1
        motion_controller_index += 2
        #print(" wrapper_name, animation_node_index, motion_controller_index ", wrapper_name, animation_node_index, motion_controller_index)

        ia_obj = object_wrapper(wrapper_name, animation_node, motion_controller_list, vector_frame_list, \
                    quaternion_frame_list)
                    
        ia_objects_list.append(ia_obj)
        parent = Object

        #print(" append {:s} {:s}".format(Object.name, Object.type))
        processed_objects_list.append(Object)
        
    else:
        parent = None

    for obj_children in Object.children:
        '''if parent is not None:
            print(" parent / child", parent.name, obj_children.name)
        else:
            print(" parent None / child", obj_children.name)'''
        if not visible_children or (visible_children and obj_children.is_visible(scene)): # MAJ 1.4.2
            animation_node_index, motion_controller_index = create_ia_objects(bpy, obj_children, parent, \
                            ia_objects_list, KeyframeCount, processed_objects_list, \
                            animation_node_index, motion_controller_index, FrameRateMultiplier, applied_translation, visible_children, scene)
     
    if parent is not None:
        # Reset object to initial frame position
        bpy.context.scene.frame_set(current_frame)

    return animation_node_index, motion_controller_index
    
         
######################################################
# EXPORT
######################################################
            
            
def ExportIA(Config):         

    # Current  model directory name
    model_directory = os.path.dirname(Config.FilePath)
    # Current model file name
    model_name = os.path.basename(Config.FilePath)
    # Name without suffix
    filename_strip = os.path.splitext(model_name)[0]
    
    myLogfile = model_directory + '\\' + filename_strip + '_ia.log'
    outfile = open(myLogfile, 'w')
    saveout = sys.stdout
    sys.stdout = outfile

    
    # Read Model specific file for Blender IA exporter
    specific_config = configparser.ConfigParser()
    specific_config_name = model_directory + '\\' + filename_strip + '_IAExpModFile2.txt'
    specific_config_found = True
    if not os.path.exists(specific_config_name):
        print(' >>>> INFO: File {:s} not found <<<<'.format(specific_config_name))
        specific_config_name = model_directory + '\\' + 'IAExpModFile2.txt'
        if not os.path.exists(specific_config_name):
            print(' >>>> INFO: File {:s} not found <<<<'.format(specific_config_name))
            specific_config_found = False
        else:
            print(' >>>> INFO: Using file {:s} <<<<'.format(specific_config_name))
    specific_config.read(specific_config_name)
    if specific_config_found:
        if not specific_config.sections():
            print(' >>>> INFO: File {:s} is empty <<<<'.format(specific_config_name))
            pass
        else:
            print_config(specific_config, specific_config_name)

    FrameRateMultiplier = 1
    try:
        myVar = specific_config.get('Miscellaneous', 'FrameRateMultiplier')
        if not myVar.isdigit():
            print(" INFO: FrameRateMultiplier {:s} ignored (non integer value found): FrameRateMultiplier reset to 1".format(myVar))
        else:
            FrameRateMultiplier = int(myVar)
            print(" INFO: FrameRateMultiplier read from configuration file: {:d}".format(FrameRateMultiplier))
    except:
        pass    
    if FrameRateMultiplier > MAXFRAMERATEMULTPLIER:
        print(" ERROR: Frame rate multiplier greater than {:d}. FrameRateMultiplier reset to 1.".format(MAXFRAMERATEMULTPLIER))
        FrameRateMultiplier = 1

    RemoveLastFrame = False
    try:
        myVar = specific_config.get('Miscellaneous', 'RemoveLastFrame')
        if not myVar.isdigit():
            print(" INFO: RemoveLastFrame {:s} ignored (non integer value found): Last frame will not be removed".format(myVar))
        else:
            if int(myVar) > 0:
                RemoveLastFrame = True
                print(" INFO: RemoveLastFrame read from configuration file: Last frame will be removed")
            else:
                print(" INFO: RemoveLastFrame read from configuration file: Last frame will not be removed")
    except:
        pass    

    # ------------------------------------------------------------------------------------------
    # Read Model specific file for Blender IGS exporter
    print("\n\n")
    scene = Config.context.scene

    specific_igs_config = configparser.ConfigParser()
    specific_igs_config_name = model_directory + '\\' + filename_strip + '_IGSExpModFile2.txt'
    specific_igs_config_found = True
    if not os.path.exists(specific_igs_config_name):
        print(' >>>> INFO: File {:s} not found <<<<'.format(specific_igs_config_name))
        specific_igs_config_name = model_directory + '\\' + 'IGSExpModFile2.txt'
        if not os.path.exists(specific_igs_config_name):
            print(' >>>> INFO: File {:s} not found <<<<'.format(specific_igs_config_name))
            specific_igs_config_found = False
        else:
            print(' >>>> INFO: Using file {:s} <<<<'.format(specific_igs_config_name))
    specific_igs_config.read(specific_igs_config_name)
    if specific_igs_config_found:
        if not specific_igs_config.sections():
            print(' >>>> INFO: File {:s} is empty <<<<'.format(specific_igs_config_name))
            pass
        else:
            print_config_section(specific_igs_config, specific_igs_config_name, 'Miscellaneous')
        
    main_object_found = False
    try:
        main_obj_name = specific_igs_config.get('Miscellaneous', 'MainObject')
        print(" INFO:  MainObject to find = {:s}".format(main_obj_name))
        for ob in scene.objects:
            if ob.name == main_obj_name:
                main_object_found = True
                main_obj = ob
                print(" MainObject found = {:s}".format(ob.name))
                break
        if not main_object_found:
            print(" ERROR:  MainObject not found")
    except:
        print(" ERROR:  MainObject not specified")
    
    applied_translation = Vector((0.0, 0.0, 0.0))
    applied_translation_ignored = False
    if main_object_found:
        try:
            center_main_object = specific_igs_config.get('Miscellaneous', 'CenterMainObject')
            center_main_object = str(center_main_object).replace(' ','') # remove spaces
            s = center_main_object.split(',')
            if len(s) != 3:
                print(" ERROR: CenterMainObject ({:s}) not set with 3 floats or 'N' separated by a comma: CenterMainObject ignored".format(center_main_object))
            else:
                applied_translation = main_obj.matrix_world.to_translation()
                for num_axis in range(3):
                    if s[num_axis] == "n" or s[num_axis] == "N":
                        applied_translation[num_axis] = 0
                    else:
                        string_error = False
                        if s[num_axis][0]=='+': s[num_axis]=s[num_axis][1:]
                        if '-' in s[num_axis] and s[num_axis][0] != '-':
                            string_error = True
                        if '+' in s[num_axis]:
                            string_error = True
                        new_s = s[num_axis].replace('-','',1)
                        if not new_s.replace('.','',1).isdigit() or string_error:
                            if num_axis == 0:
                                print(" ERROR: CenterMainObject ({:s}) ignored (non numeric value found for X)".format(center_main_object))
                            if num_axis == 1:
                                print(" ERROR: CenterMainObject ({:s}) ignored (non numeric value found for Y)".format(center_main_object))
                            if num_axis == 2:
                                print(" ERROR: CenterMainObject ({:s}) ignored (non numeric value found for Z)".format(center_main_object))
                            applied_translation_ignored = True
                        else:
                            applied_translation[num_axis] -= float(s[num_axis])
            
            if applied_translation_ignored:
                applied_translation = Vector((0.0, 0.0, 0.0))
            else:
                print(" INFO: Using CenterMainObject ({:s},{:s},{:s}), exported animation moved by X={:f} Y={:f} Z={:f} (Blender axis)".format(s[0], s[1], s[2], \
                                                                        -applied_translation.x, -applied_translation.y, -applied_translation.z))
            
        except:
            pass    
    # ------------------------------------------------------------------------------------------
            

    #Config.Verbose = True;    
    
    # Start of export
    time1 = time.clock()

    if bpy.ops.object.mode_set.poll():
        bpy.ops.object.mode_set(mode='OBJECT')

    scene = Config.context.scene

    if Config.use_selection:
#        selected_objects_list = [ob for ob in scene.objects if ob.is_visible(scene) and ob.select]
        selected_objects_list = [ob for ob in scene.objects if ob.select]
        if len(selected_objects_list) == 0:
            print(' >>>> Nothing was selected')
            restore_std_out(outfile, saveout)
            print(' >>>> Nothing was selected')
            return
    else:
        selected_objects_list = [ob for ob in scene.objects if ob.is_visible(scene)] # ob.is_visible(scene) => objects with eye in outliner not greyed out

    '''--------------------------------------------------'''
    '''Save the selected Blender objects into ia objects.'''
    '''--------------------------------------------------'''
    
    KeyframeCount = Config.context.scene.frame_end - Config.context.scene.frame_start + 1
    print('\n\n Keyframes count = {:d} (frame start={:d} / frame end={:d})'.format(KeyframeCount, Config.context.scene.frame_start, Config.context.scene.frame_end))

    print(' Frame count multipler = ', FrameRateMultiplier)
    
    if KeyframeCount == 0:
        print(" No animation found (key frame count = 0): Export stopped")
        restore_std_out(outfile, saveout)
        print(" No animation found (key frame count = 0): Export stopped")
        return {'FINISHED with error'}

    ia_objects_list = []
    processed_objects_list = []

    animation_node_index = 0
    motion_controller_index = 0
    parent = None

    if Config.use_selection:
        print(" --- Selected objects: ---")
    
    for Object in selected_objects_list:
        if Config.use_selection:
            print('     ', Object.name)
        if (Config.use_selection or not Object.parent) and has_animation_in_hierachy(Object): # If not use_selection, don't process a child before its parent
            #print(' processed_objects_list ', processed_objects_list)
            animation_node_index, motion_controller_index = create_ia_objects(bpy, Object, parent, \
                            ia_objects_list, KeyframeCount, processed_objects_list, \
                            animation_node_index, motion_controller_index, FrameRateMultiplier, applied_translation, Config.visible_children, scene)
                            
    #print(' animation_node_index = / motion_controller_index = ', animation_node_index, motion_controller_index)
    if animation_node_index == 0:
        print(" No animation found: Export stopped")
        restore_std_out(outfile, saveout)
        print(" No animation found: Export stopped")
        return {'FINISHED with error'}
    
    # Update data to deal with actual offsets in the future IA file
    if RemoveLastFrame:
        total_frames_count = (KeyframeCount - 1) * FrameRateMultiplier
    else:
        total_frames_count = (KeyframeCount - 1) * FrameRateMultiplier + 1
    start_animation_node_offset = IA_HEADER_SIZE 
    start_motion_controller_offset = IA_HEADER_SIZE + animation_node_index * IA_ANIMATIONNODE_SIZE 
    start_motion_set_offset = start_motion_controller_offset + motion_controller_index * IA_MOTIONCONTROLLER_SIZE 
    for ia_obj in ia_objects_list:
        ia_obj.animation_node.mMotionControllers = start_motion_controller_offset + ia_obj.animation_node.mMotionControllers * IA_MOTIONCONTROLLER_SIZE
        ia_obj.motion_controller[0].mKeyStream = start_motion_set_offset + ia_obj.motion_controller[0].mKeyStream * IA_MOTIONSET_SIZE * total_frames_count
        ia_obj.motion_controller[1].mKeyStream = start_motion_set_offset + ia_obj.motion_controller[1].mKeyStream * IA_MOTIONSET_SIZE * total_frames_count
        
    # Initialize Header structure
    blender_version = bpy.app.version_string
    exporter_version = EXPORTER_VERSION
    myComment = 'Blender ' + blender_version + ' - ' + exporter_version
    mTriggersTimeline = _IATriggerTimeline(0, 0)
    # mCommentField, mDataStart, mTotalFrames, mSampleRate, mTriggersTimeline, mNumAnimationNodes, mAnimationNodes, mDataEnd
    myHeader = _IAFHeader(myComment, 0, total_frames_count, FrameRateMultiplier * int(Config.context.scene.render.fps / Config.context.scene.render.fps_base), \
                            mTriggersTimeline, animation_node_index, start_animation_node_offset, 0)

    '''----------------------------------------'''
    '''Dump all the ia objects to the ia file .'''
    '''----------------------------------------'''
    
    # Open the file for writing ia data
    file = open(model_directory + '\\' + filename_strip + '.ia', 'wb')
   
    myHeader.write(file)
    if Config.Verbose:
        myHeader.dump_log()

    if Config.Verbose:
        print('================================================================================')
        print('IAfAnimationNodes             : [ {:d} ]'.format(animation_node_index))
    else:
        print('\n==========================')
        print(' --- Included objects in the animation: ---')
    for i in range(animation_node_index):
        ia_objects_list[i].animation_node.write(file)
        ia_objects_list[i].animation_node.dump_log(i+1, animation_node_index, start_animation_node_offset + i * IA_ANIMATIONNODE_SIZE, Config.Verbose)

    if Config.Verbose:
        print('================================================================================')
        print('IAfMotionControllers          : [ {:d} ]'.format(motion_controller_index))
    for i in range(animation_node_index):
        ia_objects_list[i].motion_controller[0].write(file)
        ia_objects_list[i].motion_controller[1].write(file)
        if Config.Verbose:
            ia_objects_list[i].motion_controller[0].dump_log(2*i+1, motion_controller_index, start_motion_controller_offset + 2*i * IA_MOTIONCONTROLLER_SIZE )
            ia_objects_list[i].motion_controller[1].dump_log(2*i+2, motion_controller_index, start_motion_controller_offset + (2*i+1) * IA_MOTIONCONTROLLER_SIZE )

    if Config.Verbose:
        print('================================================================================')
        print('IAfMotionSets                 : [ {:d} ]'.format(motion_controller_index))
    for i in range(animation_node_index):
        offset = ia_objects_list[i].motion_controller[0].mKeyStream
        if Config.Verbose:
            print('--------------------------------------------------------------------------------')
            print('IAfMotionSets [ {:d}  /  {:d}  ] ({:d}) :'.format(2*i+1, motion_controller_index, offset))
            print('------------------------------.')
        for frame_num in range(total_frames_count):
            ia_objects_list[i].motion_vector[frame_num].write(file)
            if Config.Verbose:
                ia_objects_list[i].motion_vector[frame_num].dump_log(frame_num)
        offset = ia_objects_list[i].motion_controller[1].mKeyStream
        if Config.Verbose:
            print('--------------------------------------------------------------------------------')
            print('IAfMotionSets [ {:d}  /  {:d}  ] ({:d}) :'.format(2*(i+1), motion_controller_index, offset))
            print('------------------------------.')
#        for frame_num in range((KeyframeCount - 1) * FrameRateMultiplier):
        for frame_num in range(total_frames_count):
            ia_objects_list[i].motion_quaternion[frame_num].write(file)
            if Config.Verbose:
                ia_objects_list[i].motion_quaternion[frame_num].dump_log(frame_num)
            
    # Close the file:
    file.close()
    
    print("\n\nIA export time: {:.02f}".format(time.clock() - time1))

#---------------------------
    restore_std_out(outfile, saveout)

    # open the log file for output to console
    '''    try:
        infile = open(myLogfile, 'r')
        for line in infile:
            print(line, end="")
    except:
        print('Error reading log file')
    else:
        infile.close()'''
 #---------------------------

    return {'FINISHED'}


class IAExporter(bpy.types.Operator):
    """Export to the IA model format (.ia) for Railworks"""

    bl_idname = "export.ia"
    bl_label = "Export ia"

    filepath = StringProperty(subtype='FILE_PATH')

    # Options
    use_selection = BoolProperty(
        name="Use selection",
        description="Enable to export only the selected objects.",
        default=True)

    visible_children = BoolProperty(
        name="Visible children",
        description="Enable to export only the visible children. Choice must be consistent with igs export.",
        default=False)

    Verbose = BoolProperty(
        name="Verbose",
        description="Detailed log file. Check the console for output",
        default=False)

    def execute(self, context):
        # Append .ia
        FilePath = bpy.path.ensure_ext(self.filepath, ".ia")

        Config = IAExporterSettings(context,
                                     FilePath,
                                     use_selection=self.use_selection,
                                     visible_children=self.visible_children,
                                     Verbose=self.Verbose)

        ExportIA(Config)
        return {"FINISHED"}

    def invoke(self, context, event):
        if not self.filepath:
            self.filepath = bpy.path.ensure_ext(bpy.data.filepath, ".ia")
        WindowManager = context.window_manager
        WindowManager.fileselect_add(self)
        return {"RUNNING_MODAL"}


def menu_func(self, context):
    self.layout.operator(IAExporter.bl_idname, text="TS20xx IA (.ia)")


def register():
    bpy.utils.register_module(__name__)

    bpy.types.INFO_MT_file_export.append(menu_func)


def unregister():
    bpy.utils.unregister_module(__name__)

    bpy.types.INFO_MT_file_export.remove(menu_func)


if __name__ == "__main__":
    register()
    