# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Thanks: 'get_autosmooth_normal' from the trainz exporter (adapted and fixed)

bl_info = {
    "name": "IGS Model Format (.igs)",
    "author": "dom107",
    "version": (1, 4, 8),
    "blender": (2, 76, 0),
    "location": "File > Export > TS20xx IGS (.igs)",
    "description": "Export IGS Model Format (.igs) for Railworks",
    "warning": "",
    "tracker_url": "",
    "category": "Import-Export"}

EXPORTER_VERSION = 'IGS exporter version 1.4.8' # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

######################################################
# Data Structures
######################################################

# vertex welding thresholds
# these are from the TRK Exporter
gVertexWeldNormalThreshold = 0.00002
gVertexWeldColourThreshold = 0.0039
gVertexWeldUVThreshold = 0.00002

IGF_MAX_RENDER_STAGES = 8
IGF_MAX_LEGACY_TEXTURE_NAME = 32
IGF_MAX_SHADER_NAME = 128
IGF_MAX_SURF_TYPE_NAME = 128
IGF_MAX_MATERIAL_COMMENT = 256
IGF_MAX_MATERIAL_NAME = 128
IGF_MAX_EFFECT_NAME = 32
IGF_MAX_LIGHTGLOW_TYPE = 32
IGF_MAX_SEQUENCE_MODE = 32
IGF_MAX_OBJECT_NAME = 32
IGF_MAX_EXPORTED_BONES = 8
IGF_MAX_HEADER_COMMENT = 64
IGF_MAX_HEADER_DESCRIPTOR = 16
IGF_MAX_CHILDREN = 24
IGF_MAX_FACE_TAG_DESCS = 256 # Each tag is of size IGF_MAX_OBJECT_NAME

IGS_HEADER_SIZE = 216
IGS_MATERIAL_SIZE = 1788
IGS_VERTEX_SIZE = 240
IGS_TRIANGLE_SIZE = 92
IGS_GEOM_SIZE = 100
IGS_GEOMLOD_SIZE = 56
IGS_INSTANCE_SIZE = 260
IGS_DATABLOCK_SIZE = 44

SIZE_OF_INT = 4

END_OK = 0
END_ERROR = -1
END_WARNING = -2

KEYWORD_OTHER = 0
KEYWORD_CUSTOM = 1

MATERIALS_SHADER_NAME = 'Materials:ShaderName'
RESERVED_OBJECT_NAMES = ['_locomotive', '_tender', '_coach', '_vehicle', '_wagon', '_carriage', '_coal', '_fuel_level_', '_freight', '_bulk', \
'_lights_fwdhead', '_lights_revhead', '_lights_fwdtail', '_lights_revtail', '_light_fwdhead', '_light_revhead', '_light_fwdtail', '_light_revtail']
RESERVED_OBJECT_NAMES_NB = ['_door', '_step', '_wh', '_bo', '_panto', '_wiper', '_primarydigits_'] # Reserved names followed by 1 or 2 digits
RESERVED_OBJECT_NAMES_1NB = ['_primarydigits_'] # Reserved names followed by 1 digit (any keyword in RESERVED_OBJECT_NAMES_1NB must be also in RESERVED_OBJECT_NAMES_NB)

# SHADERS_LIST_TO_CHECK[i][1] = expected number of slots for TS2012 / SHADERS_LIST_TO_CHECK[i][2] = related number of slots for Blender (excluding shader name slot)
SHADERS_LIST_TO_CHECK = [['BlendATexDiff',1,1] ,['TexDiff',1,1], ['TrainBumpEnv.fx',3,3], ['TrainBumpEnvMask.fx',3,3], ['TrainBumpSpec.fx',2,2], \
['TrainBumpSpecEnv.fx',3,3], ['TrainSpecEnvMask.fx',2,2], ['TrainBumpSpecEnvMask.fx',3,3], ['TrainGlassWeatherEffects.fx',4,3], ['TrainBasicObjectDiffuse.fx',1,1]]


import bpy

from bpy.props import StringProperty, EnumProperty, BoolProperty

import struct
import os

from mathutils import Vector, Matrix
from math import radians, ceil, degrees
import configparser
import sys
import time
import string

def isfloat(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True

# Container for the exporter settings
class IGSExporterSettings:
    def __init__(self,
                 context,
                 FilePath,
                 use_selection=False,
                 visible_children=False,
                 smaller_IGS_file=True,
                 Verbose=False):
        self.context = context
        self.FilePath = FilePath
        self.use_selection = use_selection
        self.visible_children = visible_children
        self.smaller_IGS_file = smaller_IGS_file
        self.Verbose = Verbose

        
class _igs_rgba_color(object):
# 16 bytes / r, g, b, alpha
    __slots__ = "r", "g", "b", "a"

    def __init__(self, r, g, b, a):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def get_size(self):
        return 16

    def write(self, file):
        file.write(struct.pack('<4f', self.r, self.g, self.b, self.a))

    def dump(self):
        return '{Colour: {:f}, {:f}, {:f}, {:f}}'.format(self.r, self.g, self.b, self.a)

    def dump_log(self):
        print("    Red                           : {:f}".format(self.r))
        print("    Green                         : {:f}".format(self.g))
        print("    Blue                          : {:f}".format(self.b))
        print("    Alpha                         : {:f}".format(self.a))

        
class _igs_vector4(object):
    __slots__ = "x", "y", "z", "w"

    def __init__(self, x, y, z, w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def get_size(self):
        return 16

    def write(self, file):
        file.write(struct.pack('<4f', self.x, self.y, self.z, self.w))

    def dump(self):
        print('{Point: {:f}, {:f}, {:f}, {:f}}'.format(self.x, self.y, self.z, self.w))

    def dump_log(self):
        print("  X                             : {:f}".format(self.x))
        print("  Y                             : {:f}".format(self.y))
        print("  Z                             : {:f}".format(self.z))
        print("  Weight                        : {:f}".format(self.w))      


class _igs_Matrix4x4(object):
    __slots__ = "M"

    def __init__(self, a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p):
        self.M = Matrix(((a,b,c,d),(e,f,g,h),(i,j,k,l),(m,n,o,p)))

    def get_matrix(self):
        return self.M

    def update_rotation(self, matrix):
        myMatrix = Matrix(((matrix[0][0], matrix[1][0],matrix[2][0]), \
        (matrix[0][1], matrix[1][1],matrix[2][1]), \
        (matrix[0][2], matrix[1][2],matrix[2][2])))

        myMatrix2 = Matrix(((self.M[0][0], self.M[1][0],self.M[2][0]), \
        (self.M[0][1], self.M[1][1],self.M[2][1]), \
        (self.M[0][2], self.M[1][2],self.M[2][2])))
        myMatrix2 = myMatrix.inverted() * myMatrix2
        for row in range (3):
            for col in range (3):
                self.M[row][col] = myMatrix2[row][col]
        
    def get_translation(self):
        return self.M[3][0], self.M[3][1], self.M[3][2]

    def substract_translation(self, m,n,o):
        self.M[3][0] -= m
        self.M[3][1] -= n
        self.M[3][2] -= o

    def get_size(self):
        return 64

    def write(self, file):
        file.write(struct.pack('<4f', self.M[0][0], self.M[0][1],self.M[0][2],self.M[0][3]))
        file.write(struct.pack('<4f', self.M[1][0], self.M[1][1],self.M[1][2],self.M[1][3]))
        file.write(struct.pack('<4f', self.M[2][0], self.M[2][1],self.M[2][2],self.M[2][3]))
        file.write(struct.pack('<4f', self.M[3][0], self.M[3][1],self.M[3][2],self.M[3][3]))

    def dump_log(self):
        print("TM -----------------------------.")
        for row in range (4):
            print("  Row{:d} ---------------------------. ".format(row))
            for col in range (4):
                print("    Column{:d}                       : {:f}".format(col, self.M[row][col]))
        
class _igs_bounding_box(object):
    __slots__ = "min", "max"

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def get_size(self):
        return 32

    def write(self, file):
        self.min.write(file)
        self.max.write(file)

    def dump(self):
        self.min.dump()
        self.max.dump()

    def dump_log(self):
        print("BoundingBox --------------------.")
        print("  LowerLeftPointX               : {:f}".format(self.min.x))
        print("  LowerLeftPointY               : {:f}".format(self.min.y))
        print("  LowerLeftPointZ               : {:f}".format(self.min.z))
        print("  LowerLeftPointScale           : {:f}".format(self.min.w))       
        print("  UpperRightPointX              : {:f}".format(self.max.x))
        print("  UpperRightPointY              : {:f}".format(self.max.y))
        print("  UpperRightPointZ              : {:f}".format(self.max.z))
        print("  UpperRightPointScale          : {:f}".format(self.max.w))      

  
class _igs_UV(object):
    __slots__ = "u", "v"

    def __init__(self, u, v):
        self.u = u
        self.v = v

    def get_size(self):
        return 8

    def write(self, file):
        file.write(struct.pack('<2f', self.u, self.v))

    def dump(self):
        return '{Vector2: {:f}, {:f}}'.format(self.u, self.v)        

    def dump_log(self):
        print("  U                             :", self.u)
        print("  V                             :", self.v)
 
 
class _IGVertexBoneBinding(object):
    __slots__ = "mBone", "mWeight"

    def __init__(self, mBone, mWeight):
        self.mBone = mBone
        self.mWeight = mWeight

    def get_size(self):
        return 8

    def write(self, file):
        file.write(struct.pack('<If', self.mBone, self.mWeight))

    def dump(self):
        return '{IGVertexBoneBinding: {:d}, {:f}}'.format(self.mBone, self.mWeight)

    def dump_log(self):
        if self.mBone == 0:        
            print("  BoneListStartOffset           : 0 NULL ==> None")
        else:
            print("  BoneListStartOffset           :", self.mBone)
        print("  Weight                        :", self.mWeight)

        
class _IGDataBlockContainer(object):
    ''' Container that details extra data blocks attached to an IG structure
        most major structures in the IG format have a data block container
    '''
    __slots__ = "NumBlocks", "offsetBlocks", "pad0", "pad1"

    def __init__(self, NumBlocks, offsetBlocks, pad0, pad1):
        self.NumBlocks = NumBlocks
        self.offsetBlocks = offsetBlocks
        assert(type(pad0) == bytes)
        self.pad0 = pad0
        self.pad1 = [pad1[0], pad1[1]] # 2 x sInt32

    def get_size(self):
        return IGF_MAX_OBJECT_NAME + 4 * SIZE_OF_INT # 48 bytes

    def write(self, file):
        block_format = "<II{:d}s2I".format(IGF_MAX_OBJECT_NAME)
        file.write(struct.pack(block_format, self.NumBlocks, self.offsetBlocks, self.pad0, self.pad1[0], self.pad1[1]))

    def dump(self):
        print(" IGDataBlockContainer NumBlocks= ", self.NumBlocks, " offsetBlocks=", self.offsetBlocks, \
        "  Pad0='", self.pad0.decode(), "' Pad1= [", self.pad1[0], ", ", self.pad1[1],"]")

    def dump_log(self):
        print("ExtraData ----------------------. ")
        print("  DataBlockCount                : {:d}".format(self.NumBlocks))
        print("  DataBlockStartOffset          : {:d}".format(self.offsetBlocks)) # (--)
        print("  Pad0                          : '{:s}'".format(self.pad0.decode()))
        print("  Pad1                          : [", self.pad1[0], ", ", self.pad1[1],"]")
        
        
class _IGFHeader(object):
# Main file header, always written first in the IG file stream
    __slots__ = "mMagic", "mVersion", "mIGDescriptor", "mCommentField", "mDataStart", "mNumInstances", "mInstances", "mNumGeometry", "mGeometry", "mNumBones", \
    "mBones", "mNumMaterials", "mMaterials", "mNumLights", "mLights", "mAmbientColour", "mNumSplines", "mSplines", "mNumGenericItems", "mGenericItems", \
    "mExtraDataNumBlocks", "mExtraOffsetBlocks", "mExtraDataPad0", "mExtraDataPad1", "DataPtr"

    def __init__(self, mCommentField, mDataStart, mNumInstances, mInstances, mNumGeometry, mGeometry, mNumBones, mBones, mNumMaterials, mMaterials, mNumLights, mLights, \
    mAmbientColour, mNumSplines, mSplines, mNumGenericItems, mGenericItems, mExtraDataNumBlocks, mExtraOffsetBlocks, DataPtr):
          self.mMagic = b'KIGF' # must be == gIGFMagicCode or not a valid IGF format file
          self.mVersion = 151 # current version of IGF file
          # generic comment / description field
          # comment can be anything, descriptor describes IG purpose (eg. 'shape', 'level', etc)
          self.mIGDescriptor = b'SHAPE'
          self.mCommentField = str.encode(mCommentField)
          self.mDataStart = mDataStart # tags beginning of file data
          self.mNumInstances = mNumInstances # number of object instances
          self.mInstances = mInstances # number of object instances
          self.mNumGeometry = mNumGeometry # number of geometry items (meshes)
          self.mGeometry = mGeometry # Start of list of geometry items
          self.mNumBones = mNumBones # Number of bones present
          self.mBones = mBones # List of bones
          self.mNumMaterials = mNumMaterials # Number of materials
          self.mMaterials = mMaterials # Start of list of materials
          self.mNumLights = mNumLights # Number of lights
          self.mLights = mLights # Start of list of lights
          self.mAmbientColour = mAmbientColour
          self.mNumSplines = mNumSplines # Number of splines
          self.mSplines = mSplines # Start of list of splines
          self.mNumGenericItems = mNumGenericItems # Number of list of generic items
          self.mGenericItems = mGenericItems # Start of list of generic items
          self.mExtraDataNumBlocks = mExtraDataNumBlocks
          self.mExtraOffsetBlocks = mExtraOffsetBlocks # 'Face tag descriptions' in IGfDataBlocks
          self.mExtraDataPad0 = b''
          self.mExtraDataPad1 = 0, 0
          self.DataPtr = DataPtr # Custom data blocks
          
    def get_size(self):
        return IGS_HEADER_SIZE

    def write(self, file):
        block_format = "<4sI{:d}s".format(IGF_MAX_HEADER_DESCRIPTOR)
        file.write(struct.pack(block_format, self.mMagic, self.mVersion, self.mIGDescriptor))
        block_format2 = "<{:d}sI".format(IGF_MAX_HEADER_COMMENT)
        file.write(struct.pack(block_format2, self.mCommentField, self.mDataStart))
        file.write(struct.pack('<4I', self.mNumInstances, self.mInstances, self.mNumGeometry, self.mGeometry))
        file.write(struct.pack('<6I4f', self.mNumBones, self.mBones, self.mNumMaterials, self.mMaterials, self.mNumLights, self.mLights, self.mAmbientColour.r,self.mAmbientColour.g,self.mAmbientColour.b,self.mAmbientColour.a))
        file.write(struct.pack('<4I', self.mNumSplines, self.mSplines, self.mNumGenericItems, self.mGenericItems))
        block_format3 = "<II{:d}sIII".format(IGF_MAX_OBJECT_NAME)
        file.write(struct.pack(block_format3, self.mExtraDataNumBlocks, self.mExtraOffsetBlocks, self.mExtraDataPad0, self.mExtraDataPad1[0], self.mExtraDataPad1[1], self.DataPtr))

    def dump_log(self, verbose):
        print("================================================================================")
        print("IG file header information (0)")
        print("--------------------------------------------------------------------------------")
        if verbose:
            print("MagicId                       : '{:s}'".format(self.mMagic.decode()))
            print("Version                       : {:d}".format(self.mVersion))
            print("Descriptor                    : '{:s}'".format(self.mIGDescriptor.decode()))
        print("CommentField                  : '{:s}'".format(self.mCommentField.decode()))
        if verbose: print("DataStart                     : {:d}".format(self.mDataStart)) # (--)
        print("ObjectCount                   : {:d}".format(self.mNumInstances))
        if verbose: print("ObjectListStartOffset         : {:d}".format(self.mInstances)) # (--)
        print("MeshCount                     : {:d}".format(self.mNumGeometry))
        if verbose:
            print("MeshListStartOffset           : {:d}".format(self.mGeometry)) # (--)
            print("BoneCount                     : {:d}".format(self.mNumBones))
            if self.mBones == 0:
                print("BoneListStartOffset           : 0 NULL ==> None")
            else:
                print("BoneListStartOffset           : {:d}".format(self.mBones))
        print("MaterialCount                 : {:d}".format(self.mNumMaterials))
        if verbose:
            print("MaterialListStartOffset       : {:d}".format(self.mMaterials)) # (--)
            print("LightCount                    : {:d}".format(self.mNumLights))
            if self.mLights == 0:
                print("LightListStartOffset          : 0 NULL ==> None")
            else:
                print("LightListStartOffset          : {:d}".format(self.mLights))
            print("AmbientColor -------------------.")
            print("  Red                           : {:f}".format(self.mAmbientColour.r))
            print("  Green                         : {:f}".format(self.mAmbientColour.g))
            print("  Blue                          : {:f}".format(self.mAmbientColour.b))
            print("  Alpha                         : {:f}".format(self.mAmbientColour.a))
            print("SplineCount                   : {:d}".format(self.mNumSplines))
            if self.mSplines == 0:
                print("SplineListStartOffset         : 0 NULL ==> None")
            else:
                print("SplineListStartOffset         : {:d}".format(self.mSplines))
            print("GenericItemCount              : {:d}".format(self.mNumGenericItems))
            if self.mGenericItems == 0:
                print("GenericItemListStartOffset    : 0 NULL ==> None")
            else:
                print("GenericItemListStartOffset    : {:d}".format(self.mGenericItems))
            print("ExtraData ----------------------. ")
            print("  DataBlockCount                : {:d}".format(self.mExtraDataNumBlocks))
            print("  DataBlockStartOffset          : {:d}".format(self.mExtraOffsetBlocks)) # (--)
            print("  Pad0                          : '{:s}'".format(self.mExtraDataPad0.decode()))
            print("  Pad1                          : [", self.mExtraDataPad1[0], ", ", self.mExtraDataPad1[1],"]")
            print("ExtraDataBlockListStartOffset : {:d}".format(self.DataPtr)) # (--) 'Face tag descriptions'


class _IGMaterials(object):
# Defines all properties that can be assigned to a surface in a polygon mesh
    __slots__ = "self", "mName", "mShaderName", "mNumRenderStages", \
    "IGRenderSmTextureName", "IGRenderSmPadding", "IGRenderSmUVChannelIndex", "IGRenderSmFilterMode", "IGRenderSmUVOperation", "IGRenderSmUVWrapU", \
    "IGRenderSmUVWrapV", "IGRenderSmMipLODBias", \
    "IGRenderSmAnimateUVs", "IGRenderSmNumFrames", "IGRenderSmFPS", "IGRenderSmScrollUVs", "IGRenderSmScrollU", "IGRenderSmScrollV", "IGRenderSmArguments", \
    "mNeverOcclude", "mSortPriority", "mZBias", "mAlphaTestMode", "mAdditiveAlphaStrength", "mZBufferMode",  \
    "mBackfaceCull", "mTwoSided", "mViewFacing", "mVisMod", "mSurfaceType", "mUseLightGlow", "mUsePulseGlow", "mLightGlowType", \
    "PulseGlowCmTargCol", "PulseGlowCmMode", "PulseGlowCmPhase", "PulseGlowCmPeriod", "PulseGlowCmAmplitude", "PulseGlowCmRptTimeMin", \
    "PulseGlowCmRptTimeMax", "PulseGlowCmExeTimeMin", "PulseGlowCmExeTimeMax", "PulseGlowCmActivationTime", "PulseGlowCmSeqMode", "PulseGlowCmVertexModify", \
    "mUnfoggable", "mForce32Bit", "mAmbient", "mDiffuse", "mEmissive", \
    "mEmissiveStrength", "mSpecular", "mSpecularPower", "mLMCastShadows", "mLMGenerateShadows", "mLMKeepVertexColours", "mLMTexelsPerMetre", "mComment", \
    "mExtraData"

    def __init__(self, mName, mShaderName, mNumRenderStages, \
    IGRenderSmTextureName, IGRenderSmPadding, IGRenderSmUVChannelIndex, IGRenderSmFilterMode, IGRenderSmUVOperation, IGRenderSmUVWrapU, \
    IGRenderSmUVWrapV, IGRenderSmMipLODBias, \
    IGRenderSmAnimateUVs, IGRenderSmNumFrames, IGRenderSmFPS, IGRenderSmScrollUVs, IGRenderSmScrollU, IGRenderSmScrollV, IGRenderSmArguments, \
    mNeverOcclude, mSortPriority, mZBias, mAlphaTestMode, mAdditiveAlphaStrength, mZBufferMode, \
    mBackfaceCull, mTwoSided, mViewFacing, mVisMod, mSurfaceType, mUseLightGlow, mUsePulseGlow, mLightGlowType, \
    PulseGlowCmTargCol, PulseGlowCmMode, PulseGlowCmPhase, PulseGlowCmPeriod, PulseGlowCmAmplitude, PulseGlowCmRptTimeMin, \
    PulseGlowCmRptTimeMax, PulseGlowCmExeTimeMin, PulseGlowCmExeTimeMax, PulseGlowCmActivationTime, PulseGlowCmSeqMode, PulseGlowCmVertexModify, \
    mUnfoggable, mForce32Bit, mAmbient, mDiffuse, mEmissive, \
    mEmissiveStrength, mSpecular, mSpecularPower, mLMCastShadows, mLMGenerateShadows, mLMKeepVertexColours, mLMTexelsPerMetre, mComment, mExtraData):

        self.mName = str.encode(mName) # original name of material from editor UI
        self.mShaderName = str.encode(mShaderName) # name of active shader
        # set of multipass rendering stages
        self.mNumRenderStages = mNumRenderStages
        # ----- cIGRenderStage -----
        # texture name index for this stage
        self.IGRenderSmTextureName = IGRenderSmTextureName
        # padding to remain compatible with original format:
        # this space normally would have the texture name directly in it
        self.IGRenderSmPadding = IGRenderSmPadding
        self.IGRenderSmUVChannelIndex = IGRenderSmUVChannelIndex # which UV channel to use
        # UV engine-side operations
        self.IGRenderSmFilterMode = IGRenderSmFilterMode
        self.IGRenderSmUVOperation = IGRenderSmUVOperation
        self.IGRenderSmUVWrapU = IGRenderSmUVWrapU
        self.IGRenderSmUVWrapV = IGRenderSmUVWrapV
        self.IGRenderSmMipLODBias = IGRenderSmMipLODBias
        # UV page animation
        self.IGRenderSmAnimateUVs = IGRenderSmAnimateUVs
        self.IGRenderSmNumFrames = IGRenderSmNumFrames # number of frames on texture page
        self.IGRenderSmFPS = IGRenderSmFPS
        # UV scrolling per-stage setup
        self.IGRenderSmScrollUVs = IGRenderSmScrollUVs
        self.IGRenderSmScrollU = IGRenderSmScrollU # rate to scroll per second
        self.IGRenderSmScrollV = IGRenderSmScrollV                
        # arbitary arguments setup in UI
        # purpose dependant on shader and operation mode
        self.IGRenderSmArguments = IGRenderSmArguments
        # ------------------------------
        # engine behavior setup
        self.mNeverOcclude = mNeverOcclude
        self.mSortPriority = mSortPriority
        self.mZBias = mZBias
        self.mAlphaTestMode = mAlphaTestMode
        self.mAdditiveAlphaStrength = mAdditiveAlphaStrength
        self.mZBufferMode = mZBufferMode
        self.mBackfaceCull = mBackfaceCull
        self.mTwoSided = mTwoSided
        self.mViewFacing = mViewFacing
        self.mVisMod = mVisMod # visible distance mod for this surface
        self.mSurfaceType = mSurfaceType # string description of surface type
        # special effects behaviors
        self.mUseLightGlow = mUseLightGlow
        self.mUsePulseGlow = mUsePulseGlow
        self.mLightGlowType = mLightGlowType
        # ----- cIGPulseGlowConfig -----
        self.PulseGlowCmTargCol = PulseGlowCmTargCol
        self.PulseGlowCmMode = PulseGlowCmMode
        # if in pulse mode...
        self.PulseGlowCmPhase = PulseGlowCmPhase
        self.PulseGlowCmPeriod = PulseGlowCmPeriod
        self.PulseGlowCmAmplitude = PulseGlowCmAmplitude
        # if in random mode...
        self.PulseGlowCmRptTimeMin = PulseGlowCmRptTimeMin
        self.PulseGlowCmRptTimeMax = PulseGlowCmRptTimeMax
        self.PulseGlowCmExeTimeMin = PulseGlowCmExeTimeMin
        self.PulseGlowCmExeTimeMax = PulseGlowCmExeTimeMax
        self.PulseGlowCmActivationTime = PulseGlowCmActivationTime
        # if in sequenced mode...
        self.PulseGlowCmSeqMode = PulseGlowCmSeqMode
        self.PulseGlowCmVertexModify = PulseGlowCmVertexModify # if false, texture modify
        # ------------------------------
        # behaviors promoted from comment field toggles
        self.mUnfoggable = mUnfoggable
        self.mForce32Bit = mForce32Bit
        # colouring and behavior of material surface
        self.mAmbient = mAmbient
        self.mDiffuse = mDiffuse
        self.mEmissive = mEmissive
        self.mEmissiveStrength = mEmissiveStrength
        self.mSpecular = mSpecular
        self.mSpecularPower = mSpecularPower
        # lightmap generation support
        self.mLMCastShadows = mLMCastShadows
        self.mLMGenerateShadows = mLMGenerateShadows
        self.mLMKeepVertexColours = mLMKeepVertexColours
        self.mLMTexelsPerMetre = mLMTexelsPerMetre
        self.mComment = str.encode(mComment) # generic 'comment' string
        self.mExtraData = mExtraData
          
    def get_size(self):
        return IGS_MATERIAL_SIZE

    def write(self, file):
        block_format = "<{:d}s".format(IGF_MAX_MATERIAL_NAME)
        file.write(struct.pack(block_format, self.mName))
        block_format2 = "<{:d}s".format(IGF_MAX_SHADER_NAME)
        file.write(struct.pack(block_format, self.mShaderName))
        file.write(struct.pack('<I', self.mNumRenderStages))
        block_format3 = "<I{:d}sII".format(IGF_MAX_LEGACY_TEXTURE_NAME-4)
        block_format4 = "<4I{:d}sI".format(IGF_MAX_SURF_TYPE_NAME)
        block_format5 = "<I{:d}s".format(IGF_MAX_LIGHTGLOW_TYPE)
        block_format6 = "<2f{:d}sI".format(IGF_MAX_SEQUENCE_MODE)
        block_format7 = "<f3If{:d}s".format(IGF_MAX_MATERIAL_COMMENT)
        block_format8 = "<II{:d}sIII".format(IGF_MAX_OBJECT_NAME)
        for i in range(0, self.mNumRenderStages):  # total size of cIGRenderStage = IGF_MAX_RENDER_STAGES * 104 bytes
            file.write(struct.pack(block_format3, self.IGRenderSmTextureName[i], self.IGRenderSmPadding[i], self.IGRenderSmUVChannelIndex[i], self.IGRenderSmFilterMode[i]))
            file.write(struct.pack('<3If', self.IGRenderSmUVOperation[i], self.IGRenderSmUVWrapU[i], self.IGRenderSmUVWrapV[i], self.IGRenderSmMipLODBias[i]))
            file.write(struct.pack('<4I', self.IGRenderSmAnimateUVs[i], self.IGRenderSmNumFrames[i], self.IGRenderSmFPS[i], self.IGRenderSmScrollUVs[i]))
            UVarg = self.IGRenderSmArguments[i]
            file.write(struct.pack('<ff6f', self.IGRenderSmScrollU[i], self.IGRenderSmScrollV[i], UVarg[0], UVarg[1], UVarg[2], UVarg[3], UVarg[4], UVarg[5]))
        for i in range(self.mNumRenderStages, IGF_MAX_RENDER_STAGES):
            for j in range(1, 105):
                file.write(b'\x00') # 104 null bytes
        file.write(struct.pack("<6I", self.mNeverOcclude ,self.mSortPriority, self.mZBias, self.mAlphaTestMode, self.mAdditiveAlphaStrength, self.mZBufferMode))
        file.write(struct.pack(block_format4, self.mBackfaceCull, self.mTwoSided, self.mViewFacing, self.mVisMod, self.mSurfaceType, self.mUseLightGlow))
        file.write(struct.pack(block_format5, self.mUsePulseGlow, self.mLightGlowType))
        self.PulseGlowCmTargCol.write(file)
        file.write(struct.pack("<If", self.PulseGlowCmMode, self.PulseGlowCmPhase))
        file.write(struct.pack("<5f", self.PulseGlowCmPeriod, self.PulseGlowCmAmplitude, self.PulseGlowCmRptTimeMin, self.PulseGlowCmRptTimeMax, self.PulseGlowCmExeTimeMin))
        file.write(struct.pack(block_format6, self.PulseGlowCmExeTimeMax, self.PulseGlowCmActivationTime, self.PulseGlowCmSeqMode, self.PulseGlowCmVertexModify))
        file.write(struct.pack("<2I", self.mUnfoggable, self.mForce32Bit))
        self.mAmbient.write(file)
        self.mDiffuse.write(file)
        self.mEmissive.write(file)
        file.write(struct.pack("<f", self.mEmissiveStrength))
        self.mSpecular.write(file)
        file.write(struct.pack(block_format7, self.mSpecularPower, self.mLMCastShadows, self.mLMGenerateShadows, self.mLMKeepVertexColours, self.mLMTexelsPerMetre, self.mComment))
        self.mExtraData.write(file)
        
    def dump_log(self, num_mat, nb_mat, mat_offset, texture_names, detailed, verbose):
        if detailed:
            print("--------------------------------------------------------------------------------")
            if verbose:
                print("IGfMaterials [ {:d} / {:d} ] ({:d}) : '{:s}'".format(num_mat, nb_mat, mat_offset, self.mName.decode()))
            else:
                print("Material main parameters [ {:d} / {:d} ] : '{:s}'".format(num_mat, nb_mat, self.mName.decode()))
            print("------------------------------.")
            print("Name                          : '{:s}'".format(self.mName.decode()))
            print("ShaderName                    : '{:s}'".format(self.mShaderName.decode()))
            print("RenderStageCount              : {:d}".format(self.mNumRenderStages))
            for i in range(0, self.mNumRenderStages):
                print("RenderStage{:d} -------------------.".format(i))
                if verbose:
                    print("  TextureNameStartOffset        : {:d}".format(self.IGRenderSmTextureName[i])) # (--)
                    print("  Padding                       : '{:s}'".format(self.IGRenderSmPadding[i].decode()))
                print("  UVChannelIndex                : {:d}".format(self.IGRenderSmUVChannelIndex[i]))
                print("  FilterMode                    : {:d}".format(self.IGRenderSmFilterMode[i]))
                if verbose:
                    print("  UVOperation                   : {:d}".format(self.IGRenderSmUVOperation[i]))
                    print("  UVWrapU                       : {:d}".format(self.IGRenderSmUVWrapU[i]))
                    print("  UVWrapV                       : {:d}".format(self.IGRenderSmUVWrapV[i]))
                print("  MipLODBias                    : {:f}".format(self.IGRenderSmMipLODBias[i]))
                print("  AnimateUVs                    : {:d}".format(self.IGRenderSmAnimateUVs[i]))
                print("  NumFrames                     : {:d}".format(self.IGRenderSmNumFrames[i]))
                print("  FPS                           : {:d}".format(self.IGRenderSmFPS[i]))
                if verbose:
                    print("  ScrollUVs                     : {:d}".format(self.IGRenderSmScrollUVs[i]))
                    print("  ScrollU                       : {:f}".format(self.IGRenderSmScrollU[i]))
                    print("  ScrollV                       : {:f}".format(self.IGRenderSmScrollV[i]))
                UVarg = self.IGRenderSmArguments[i]
                print("  UVArguments                   : [{:f},{:f},{:f},{:f},{:f},{:f}]".format(UVarg[0], UVarg[1], UVarg[2], UVarg[3], UVarg[4], UVarg[5]))
            if verbose: print("NeverOccludeFlag              : {:d}".format(self.mNeverOcclude))
            if verbose: print("SortPriority                  : {:d}".format(self.mSortPriority))
            print("ZBias                         : {:d}".format(self.mZBias))
            print("AlphaTestMode                 : {:d}".format(self.mAlphaTestMode))
            if verbose: print("AdditiveAlphaStrength         : {:d}".format(self.mAdditiveAlphaStrength))
            print("ZBufferMode                   : {:d}".format(self.mZBufferMode))
            print("BackfaceCull                  : {:d}".format(self.mBackfaceCull))
            print("TwoSided                      : {:d}".format(self.mTwoSided))
            print("ViewFacing                  : {:d}".format(self.mViewFacing))
            if verbose:
                print("VisibleDistanceMod            : {:d}".format(self.mVisMod))
                print("SurfaceTypeName               : '{:s}'".format(self.mSurfaceType.decode()))
                print("UseLightGlowFlag              : {:d}".format(self.mUseLightGlow))
                print("UsePulseGlowFlag              : {:d}".format(self.mUsePulseGlow))
                print("LightGlowTypeName             : '{:s}'".format(self.mLightGlowType.decode()))
                print("PulseGlowConfig ----------------.")
                print("  TargColor ----------------------.")
                self.PulseGlowCmTargCol.dump_log()
                print("  Mode                          : {:d}".format(self.PulseGlowCmMode))
                print("  Phase                         : {:f}".format(self.PulseGlowCmPhase))
                print("  Period                        : {:f}".format(self.PulseGlowCmPeriod))
                print("  Amplitude                     : {:f}".format(self.PulseGlowCmAmplitude))
                print("  RptTimeMin                    : {:f}".format(self.PulseGlowCmRptTimeMin))
                print("  RptTimeMax                    : {:f}".format(self.PulseGlowCmRptTimeMax))
                print("  ExeTimeMin                    : {:f}".format(self.PulseGlowCmExeTimeMin))
                print("  ExeTimeMax                    : {:f}".format(self.PulseGlowCmExeTimeMax))
                print("  ActivationTime                : {:f}".format(self.PulseGlowCmActivationTime))
                print("  SeqMode                       : '{:s}'".format(self.PulseGlowCmSeqMode.decode()))
                print("  VertexModify                  : {:d}".format(self.PulseGlowCmVertexModify))
                print("UnfoggableFlag                : {:d}".format(self.mUnfoggable))
                print("Force32BitFlag                : {:d}".format(self.mForce32Bit))
            print("AmbientColor -------------------.")
            self.mAmbient.dump_log()
            print("DiffuseColor -------------------.")
            self.mDiffuse.dump_log()
            print("EmissiveColor -------------------.")
            self.mEmissive.dump_log()
            print("EmissiveStrength              :  {:f}".format(self.mEmissiveStrength))
            print("SpecularColor ------------------.")
            self.mSpecular.dump_log()
            print("SpecularPower                 : {:f}".format(self.mSpecularPower))
            if verbose:
                print("LightMapCastShadows           : {:d}".format(self.mLMCastShadows))
                print("LightMapGenerateShadows       : {:d}".format(self.mLMGenerateShadows))
                print("LightMapKeepVertexColours     : {:d}".format(self.mLMKeepVertexColours))
                print("LightMapTexelsPerMetre        : {:f}".format(self.mLMTexelsPerMetre))
                print("MaterialComment               : '{:s}'".format(self.mComment.decode()))
                self.mExtraData.dump_log()
        else:
            UVarg = self.IGRenderSmArguments[0]
#            print(" '{:s}' | Shader : '{:s}' | First texture : {:s} / UVArguments: [{:f},{:f},{:f},{:f},{:f},{:f}]".format(self.mName.decode(), self.mShaderName.decode(), texture_names.get_texture_name(self.IGRenderSmTextureName[0]-1), UVarg[0], UVarg[1], UVarg[2], UVarg[3], UVarg[4], UVarg[5]))
            print(" {:<20} | {:<30} | {:<40} / [{:f},{:f},{:f},{:f},{:f},{:f}]".format(self.mName.decode(), self.mShaderName.decode(), texture_names.get_texture_name(self.IGRenderSmTextureName[0]-1), UVarg[0], UVarg[1], UVarg[2], UVarg[3], UVarg[4], UVarg[5]))

            
class _IGVertex(object):
# Defines single vertex in a polygon mesh
    __slots__ = "self", "mPoint", "mNormal", "mNumValidUVs", "mUV", "mColour", "mMaterial", "mNumBones", "mBoneBinding", "mReserved", "mExtraData"

    def __init__(self, mPoint, mNormal, mNumValidUVs, mUV, mColour, mMaterial, mNumBones, mBoneBinding, mReserved , mExtraData):
        self.mPoint = mPoint
        self.mNormal = mNormal
        self.mNumValidUVs = mNumValidUVs
        self.mUV = mUV
        self.mColour = mColour
        self.mMaterial = mMaterial
        self.mNumBones = mNumBones
        self.mBoneBinding = mBoneBinding
        self.mReserved = mReserved
        self.mExtraData = mExtraData
         
    def get_size(self):
        return IGS_VERTEX_SIZE

    def write(self, file):
        self.mPoint.write(file)
        self.mNormal.write(file)
        file.write(struct.pack('I', self.mNumValidUVs))
        for i in range(IGF_MAX_RENDER_STAGES):
            if i < self.mNumValidUVs:
                self.mUV[i].write(file)
            else:
                file.write(b'\x00\x00\x00\x00\x00\x00\x00\x00')
        self.mColour.write(file)
        file.write(struct.pack("<2I", self.mMaterial, self.mNumBones))
        for i in range(0, IGF_MAX_EXPORTED_BONES): # cIGVertexBoneBinding
            if i < self.mNumBones:
                self.mBoneBinding[i].write(file)
            else:
                file.write(b'\x00\x00\x00\x00\x00\x00\x00\x00')
        file.write(struct.pack("<I", self.mReserved))
        self.mExtraData.write(file)

    def dump_log(self, num_vert, nb_vert, vert_offset):
        print("--------------------------------------------------------------------------------")
        print("IGfVerts [ {:d} / {:d} ] ({:d})".format(num_vert, nb_vert, vert_offset))
        print("------------------------------.")
        print("Point --------------------------.")
        self.mPoint.dump_log()
        print("Normal -------------------------.")
        self.mNormal.dump_log()
        print("ValidUVsCount                 : {:d}".format(self.mNumValidUVs))
        for i in range(IGF_MAX_RENDER_STAGES):
            print("UV0{:d} ---------------------------.".format(i))
            if i < self.mNumValidUVs:
                self.mUV[i].dump_log()
            else:
                print("  U                             : 0.0")
                print("  V                             : 0.0")
        print("Color --------------------------.")
        self.mColour.dump_log()
        print("MaterialStartOffset           : {:d}".format(self.mMaterial))
        print("BoneCount                     : {:d}".format(self.mNumBones))
        for i in range(IGF_MAX_EXPORTED_BONES):
            print("BoneBinding0{:d} ------------------.".format(i))
            if i < len(self.mBoneBinding):
                self.mBoneBinding[i].dump_log()
            else:
                print("  BoneListStartOffset           :  0 NULL ==> None")
                print("  Weight                        : 0.0")
        print("Reserved                      : {:d}".format(self.mReserved))
        self.mExtraData.dump_log()


class _IGTriangle(object):
# Defines a triangle representing a face in a polygon mesh
    __slots__ = "self", "mVertex", "mNormal", "mHiddenEdges", "mMaterial", "mSMGroup", "mTagId", "mExtraData", "vertices_count_start"

    def __init__(self, mVertex, mNormal, mHiddenEdges, mMaterial, mSMGroup, mTagId, mExtraData, vertices_count_start): # three vertices making up this triangle
        self.mVertex = [mVertex[0] + vertices_count_start, mVertex[1] + vertices_count_start, mVertex[2] + vertices_count_start] # Vertex position in object space
        self.mNormal = mNormal # face normal
        self.mHiddenEdges = mHiddenEdges # edge visibility flags (EDGE_V0_V1_HIDDEN, etc)
        self.mMaterial = mMaterial # assigned material
        self.mSMGroup = mSMGroup # smoothing group ID from editor (if applicable - eg. MAX)
        self.mTagId = mTagId # per-face tag
        self.mExtraData = mExtraData # per-face extra data        
        
    def get_size(self):
        return IGS_TRIANGLE_SIZE

    def write(self, file):
        file.write(struct.pack('<3I', self.mVertex[0], self.mVertex[1], self.mVertex[2]))
        self.mNormal.write(file)
        file.write(struct.pack('<4I', self.mHiddenEdges, self.mMaterial, self.mSMGroup, self.mTagId))
        self.mExtraData.write(file)

    def dump_log(self, num_tri, nb_tri, tri_offset):
        print("--------------------------------------------------------------------------------")
        print("IGfTriangles [ {:d} / {:d} ] ({:d})".format(num_tri, nb_tri, tri_offset))
        print("------------------------------.")
        print("Vertex0                       : {:d}".format(self.mVertex[0])) # (--)
        print("Vertex1                       : {:d}".format(self.mVertex[1])) # (--)
        print("Vertex2                       : {:d}".format(self.mVertex[2])) # (--)
        print("Normal -------------------------.")
        self.mNormal.dump_log()
        print("HiddenEdges                   : {:d}".format(self.mHiddenEdges))
        print("MaterialStartOffset           : {:d}".format(self.mMaterial)) # (--)
        print("SMGroup                       : {:d}".format(self.mSMGroup))
        print("TagID                         : {:d}".format(self.mTagId))
        self.mExtraData.dump_log()

class _IGGeom(object):
# a geometry node encapsulates all data required to define a complete polygon mesh
# note that this data may be referenced and shared amongst various instances in the scene
    __slots__ = "self", "mIsSkinned", "mNumVertices", "mVertices", "mNumTriangles", "mTriangles", "mBoundingBox", "mExtraData"

    def __init__(self, mIsSkinned, mNumVertices, mVertices, mNumTriangles, mTriangles, mBoundingBox, mExtraData):
        self.mIsSkinned = mIsSkinned # has skinning information in the vertices  
        self.mNumVertices = mNumVertices # number of vertices in mVertexList  
        self.mVertices = mVertices # array of vertices used by this node  
        self.mNumTriangles = mNumTriangles # number of triangles in mTriangleList  
        self.mTriangles = mTriangles # triangles forming the surface of this mesh  
        self.mBoundingBox = mBoundingBox # bounding volume for all node verts  
        self.mExtraData = mExtraData # per-node extra data  
        
    def get_size(self):
        return IGS_GEOM_SIZE

    def write(self, file):
        file.write(struct.pack('<5I', self.mIsSkinned, self.mNumVertices, self.mVertices, self.mNumTriangles, self.mTriangles))
        self.mBoundingBox.write(file)
        self.mExtraData.write(file)

    def dump_log(self, num_geom, nb_geom, geom_offset):
        print("--------------------------------------------------------------------------------")
        print("IGfMeshes [ {:d} / {:d} ] ({:d})".format(num_geom, nb_geom, geom_offset))
        print("------------------------------.")
        print("IsSkinned                     : {:d}".format(self.mIsSkinned))
        print("VerticesCount                 : {:d}".format(self.mNumVertices))
        print("VerticesStartOffset           : {:d}".format(self.mVertices)) # (--)
        print("TrianglesCount                : {:d}".format(self.mNumTriangles))
        print("TrianglesStartOffset          : {:d}".format(self.mTriangles)) # (--)
        self.mBoundingBox.dump_log()
        self.mExtraData.dump_log()


class _IGGeometryLOD(object):

    __slots__ = "self", "VisibleDistance", "pGeometry", "mExtraData"

    def __init__(self, VisibleDistance, pGeometry, mExtraData):
        self.VisibleDistance = VisibleDistance # distance that LOD is visible
        self.pGeometry = pGeometry # mesh
        self.mExtraData = mExtraData # lod-level extra data
        
    def get_size(self):
        return IGS_GEOMLOD_SIZE

    def write(self, file):
        file.write(struct.pack('<fI', float(self.VisibleDistance),self.pGeometry))
        self.mExtraData.write(file)
        
    def dump_log(self, num_geomLOD, nb_geomLOD, geomLOD_offset):
        print("--------------------------------------------------------------------------------")
        print("IGfMeshLODs [ {:d} / {:d} ] ({:d})".format(num_geomLOD, nb_geomLOD, geomLOD_offset))
        print("------------------------------.")
        print("VisibleDistance               : {:f}".format(float(self.VisibleDistance)))
        print("MeshListStartOffset           : {:d}".format(self.pGeometry)) # (--)
        self.mExtraData.dump_log()


class _IGInstance(object):
# an instance binds translational information to a polygon mesh / IGfObject
    __slots__ = "self", "mName", "mNameUID", "mNumLODs", "mGeometry", "mTM", "mParent", "mNumChildren", "mChildren", \
    "mExtraData", "mInitialTM", "parent_group", "child_groups_list"

    def __init__(self, mName, mNameUID, mNumLODs, mGeometry, mTM, mParent, mNumChildren, mChildren, \
    mExtraData, mInitialTM, parent_group, child_groups_list):
        self.mName = str.encode(mName)  # original name of object in editor 
        self.mNameUID = mNameUID # unique ID for this object name, used for fast matching or hashing 
        self.mNumLODs = mNumLODs # number of LODs stored in mGeometry 
        self.mGeometry = mGeometry # pointers back to geometry LODs 
        self.mTM = mTM  # transformation (Matrix 4 x 4) for igs object
        # hierarchy management  
        self.mParent = mParent # parent node, None if root of hierarchy 
        self.mNumChildren = mNumChildren # number of children in mChildren list 
        self.mChildren = mChildren  # 1 or more children instances 
        self.mExtraData = mExtraData # per-instance extra data
        #
        # (initial) transformation (Matrix 4 x 4)
        self.mInitialTM = mInitialTM
        self.parent_group= parent_group # None if root of hierarchy (from group_wrapper)
        self.child_groups_list = child_groups_list  # 1 or more children instances (from group_wrapper)
        
    def get_size(self):
        return IGS_INSTANCE_SIZE

    def write(self, file):
        block_format = "<{:d}s".format(IGF_MAX_OBJECT_NAME)
        file.write(struct.pack(block_format, self.mName))
        file.write(struct.pack('<3I', self.mNameUID, self.mNumLODs, self.mGeometry))
        self.mTM.write(file)
        file.write(struct.pack('<2I', self.mParent, self.mNumChildren))
        for i in range(0, IGF_MAX_CHILDREN):
            if i < len(self.mChildren):
                file.write(struct.pack('<I', self.mChildren[i]))
            else:
                file.write(b'\x00\x00\x00\x00')
        self.mExtraData.write(file)

    def dump_log(self, num_Obj, nb_Obj, Obj_offset):
        print("--------------------------------------------------------------------------------")
        print("IGfObjects [ {:d} / {:d} ] ({:d})".format(num_Obj, nb_Obj, Obj_offset))
        print("------------------------------.")
        print("Name                          : '{:s}'".format(self.mName.decode()))
        print("NameUniqueID                  : {:d}".format(self.mNameUID))
        print("MeshLODCount                  : {:d}".format(self.mNumLODs))
        print("MeshLODListStartOffset        : {:d}".format(self.mGeometry)) # (--)
        self.mTM.dump_log()
        print("ParentStartOffset             : {:d}".format(self.mParent)) # (--)
        print("ChildrenCount                 : {:d}".format(self.mNumChildren))
        for i in range(0, IGF_MAX_CHILDREN):
            if i < len(self.mChildren):
                if self.mChildren[i] == 0:
                    print("ChildStartOffset{:02d}            : 0 NULL ==> None".format(i))
                else:
                    print("ChildStartOffset{:02d}            : {:d}".format(i, self.mChildren[i]))
            else:
                print("ChildStartOffset{:02d}            : 0 NULL ==> None".format(i))
        self.mExtraData.dump_log()

        
class _IGDataBlock(object):

    __slots__ = "self", "mName", "mID", "mData", "mSize"

    def __init__(self, mName, mID, mData, mSize):
        self.mName = str.encode(mName)  # name of data block 
        self.mID = mID  # unique id for this type 
        self.mData = mData  # pointer to arbitrary data 
        self.mSize = mSize  # size of data  
        
    def get_size(self):
        return IGS_DATABLOCK_SIZE

    def write(self, file):
        block_format = "<{:d}s".format(IGF_MAX_OBJECT_NAME)
        file.write(struct.pack(block_format, self.mName))
        file.write(struct.pack('<3I', self.mID,self.mData, self.mSize))
        
    def dump_log(self, num_dataBlock, nb_dataBlock, dataBlock_offset):
        print("--------------------------------------------------------------------------------")
        print("IGfDataBlocks [ {:d} / {:d} ] ({:d})".format(num_dataBlock, nb_dataBlock, dataBlock_offset))
        print("------------------------------.")
        print("Name                          : '{:s}'".format(self.mName.decode()))
        print("UniqueID                      : {:d}".format(self.mID))
        print("DataStartOffset               : {:d}".format(self.mData)) # (--)
        print("DataSize                      : {:d}".format(self.mSize))
        print("--------------------------------------------------------------------------------")

        
class _IGFaceTagDescription(object):

    __slots__ = "self", "FaceTagDescription"

    def __init__(self, ListOfTags):
        #self.FaceTagDescription = []
        #for i in range(0, len(ListOfTags)):
        #    self.FaceTagDescription.append(ListOfTags[i])
        self.FaceTagDescription = list(ListOfTags)
          
    def get_size(self):
        return IGF_MAX_FACE_TAG_DESCS * IGF_MAX_OBJECT_NAME

    def write(self, file):
        block_format = "<{:d}s".format(IGF_MAX_OBJECT_NAME)
        for i in range(0, len(self.FaceTagDescription)):
            file.write(struct.pack(block_format, self.FaceTagDescription[i]))
        for i in range(len(self.FaceTagDescription), IGF_MAX_FACE_TAG_DESCS):
            for j in range(0, IGF_MAX_OBJECT_NAME):
                file.write(b'\x00')
       
    def dump_log(self, dataBlock_offset):
        print('IGfFaceTagDescriptions        : [ 1 ]')
        print('--------------------------------------------------------------------------------')
        print("IGfFaceTagDescriptions [ 1 / 1 ] ({:d})".format(dataBlock_offset))
        print('------------------------------.')
        print('FaceTagDescriptionList ---------.') 
        block_format = "<%ds" % IGF_MAX_OBJECT_NAME
        for i in range(0, len(self.FaceTagDescription)):
            print("  FaceTagDescription[{:d}]        : '{:s}'".format((i+1), self.FaceTagDescription[i].decode()))
        for i in range(len(self.FaceTagDescription), IGF_MAX_FACE_TAG_DESCS):
            print("  FaceTagDescription[{:d}]        :".format(i+1))
        
        
class _IGTextureNames(object): # IGTextureNames
    ''' Container for texture names with no length limit (namespaced)
    '''
    __slots__ = "SizeStringtable", "mNumStrings", "ListOfStrings", "TextureNameStartOffset"

    def __init__(self, ListOfStrings):
        self.SizeStringtable = 0 # total length of string table
        self.mNumStrings = len(ListOfStrings) # num of strings in string table
        # Compute indices into string table (TextureNameStartOffset)
        self.TextureNameStartOffset = []
        for i in range(0, self.mNumStrings):
            if i == 0:
                self.TextureNameStartOffset.append(0)
                PrevTextureNameStartOffset = 0
            else:
                self.TextureNameStartOffset.append(PrevTextureNameStartOffset + len(ListOfStrings[i-1]) + 1)
                PrevTextureNameStartOffset = self.TextureNameStartOffset[i]
            self.SizeStringtable = self.SizeStringtable + len(ListOfStrings[i]) + 1
        # Strings table begins after indices
        for i in range(0, self.mNumStrings):
            self.ListOfStrings = list(ListOfStrings)

    def get_size(self):
        return self.SizeStringtable 

    def write(self, file):
        file.write(struct.pack('<II', self.SizeStringtable, self.mNumStrings))
        for i in range(0, self.mNumStrings):
            file.write(struct.pack('<I', self.TextureNameStartOffset[i]))
        for i in range(0, self.mNumStrings):
            block_format = "<{:d}s".format(len(self.ListOfStrings[i]))
            file.write(struct.pack(block_format, str.encode(self.ListOfStrings[i])))
            file.write(b'\x00')
        
    def dump_log(self, num_textureNames, nb_textureNames, textureNames_offset):
        print("IGTextureNames               : [ 1 ]")
        print("--------------------------------------------------------------------------------")
        print("IGfTextureNames [ {:d} / {:d} ] ({:d})".format(num_textureNames, nb_textureNames, textureNames_offset))
        print("------------------------------.")
        print("TextureNameListLength         : {:d}".format(self.SizeStringtable))
        print("TextureNameCount              : {:d}".format(self.mNumStrings))
        print("--------------------------------------------------------------------------------")
        print("TextureNameStartOffsetList -----.")
        for i in range(0, self.mNumStrings):
            print("  TextureNameStartOffset[{:d}]     : {:d} (TextureName[{:d}])".format(i+1, self.TextureNameStartOffset[i], i+1))
        print("TextureNameList ----------------.")
        for i in range(0, self.mNumStrings):
            print("  TextureName[{:d}]                : '{:s}'".format(i+1, self.ListOfStrings[i]))
        
    def get_texture_name(self, i):
        return self.ListOfStrings[i]

# Add a texture file name to the textures list
def add_texture_to_list(target_textures_directory, texture_file_name, textures_list, specific_config): # MAJ_1_4_0b
    
    i = texture_file_name.rindex('.')
    # Remove suffix
    texture_file_name = texture_file_name[:i]
    
    # Process texture file substitution   # MAJ_1_4_0b
    if specific_config.has_section('Textures:TextureName'):
        if specific_config.has_option('Textures:TextureName', texture_file_name):
            texture_file_name = specific_config.get('Textures:TextureName', texture_file_name)
    if specific_config.has_section('Textures:TextureNameStringToReplace'):
        for string_to_replace in specific_config['Textures:TextureNameStringToReplace']:
            texture_file_name_lower = texture_file_name.lower()
            if texture_file_name_lower.find(string_to_replace) >= 0:
                replacement_string = specific_config.get('Textures:TextureNameStringToReplace', string_to_replace)
                texture_file_name = texture_file_name_lower.replace(string_to_replace, replacement_string)
    # =================================================

    # Add target textures directory
    if len(target_textures_directory) > 0:
        texture_file_name = target_textures_directory + '\\' + texture_file_name
        
    textures_list.append(texture_file_name)
    return len(textures_list)

# Return reserved name found in an object name
def has_reserved_name(name, custom_keywords):

    for reserved_name in RESERVED_OBJECT_NAMES:
        if name.find(reserved_name) >= 0:
            return reserved_name, KEYWORD_OTHER

    for reserved_name in RESERVED_OBJECT_NAMES_NB:
        i = name.find(reserved_name)
        #print(" 0>> ", name, len(name), i, len(name) - len(reserved_name) - i)
        if i != -1:
            # A keyword was found
            if len(name) - len(reserved_name) - i >= 2 and reserved_name not in RESERVED_OBJECT_NAMES_1NB:
                #print(" 1>> ", name, len(name), i, name[i + len(reserved_name)], name[i + 1 + len(reserved_name)])
                if  i >= 0 and \
                name[i + len(reserved_name)].isdigit() and name[i + 1 + len(reserved_name)].isdigit():
                    # Check for wh after bo?? (bo??wh??)
                    k = name.find('_bo')
                    if k >= 0:
                        l = name.find('wh', k + len('_bo') + 2, k + len('_bo') + 4)
                    if  k >= 0 and l >=0 and \
                    name[l + 2].isdigit() and name[l + 3].isdigit() and len(name)-k >= 9: # 1_0100_bo01wh01, 1_0100_bo01wh01G, ...
                        # bo??wh??
                        return name[k:l+4], KEYWORD_OTHER
                    else:
                        if len(name)- i == len(reserved_name) + 2:
                        # bo??, for example (then, bo01 or bo02, ... is returned)
                            return name[i:i + 2 + len(reserved_name)], KEYWORD_OTHER
                    return reserved_name, KEYWORD_OTHER
            else:
                for reserved_name in RESERVED_OBJECT_NAMES_1NB: # MAJ_1_4_1
                    i = name.find(reserved_name)
                    #print(" 3>> ", reserved_name, name, len(name), i, len(name)- i, len(reserved_name) + 1)
                    # keyword with 1 or 2 digits? (then, for example, _primarydigits_ followed by the digit is returned)
                    # Ex: 1_0032_primarydigits_3 or 1_0032_primarydigits_10 or 1_0032_primarydigits_3.001 or 1_0032_primarydigits_10.001
                    if ( ((len(name) == i + len(reserved_name) + 1) and name[i + len(reserved_name)].isdigit()) or \
                         ((len(name) > i + len(reserved_name) + 1) and name[i + len(reserved_name)].isdigit() and not name[i + len(reserved_name) + 1].isdigit()) ):
                        return name[i:i + 1 + len(reserved_name)], KEYWORD_OTHER
                    else:
                        if ( ((len(name) == i + len(reserved_name) + 2) and name[i + len(reserved_name)].isdigit() and name[i + len(reserved_name) + 1].isdigit()) or \
                             ((len(name) > i + len(reserved_name) + 2) and name[i + len(reserved_name)].isdigit() and name[i + len(reserved_name) + 1].isdigit() and not name[i + len(reserved_name) + 2].isdigit()) ):
                            return name[i:i + 2 + len(reserved_name)], KEYWORD_OTHER
                        
    for reserved_name in custom_keywords:
        i = name.find(reserved_name)
        if  i >= 0:
            return reserved_name, KEYWORD_CUSTOM
                        
    return None, None


def add_igs_material(bpy, material, default_config, specific_config, textures_list):
    '''Add an igs material description out of a blender material.'''

    diffuse = []
    material_found = False
    num_slot = 0
    add_face_tag_description = False
    try:
        target_textures_directory = specific_config.get('Miscellaneous', 'TargetTexturesDirectory')
    except:
        target_textures_directory = ''
        pass

    for num_s, s in enumerate(material.texture_slots):
        #print(" >>> NEW SLOT")
        if material.use_textures[num_s] and s: # Check is slot is enabled and not empty
            #print(" >>> NEW non empty SLOT")
            if num_slot == 0:
            # Read shader name
                try:
                    mShaderName = default_config.get(MATERIALS_SHADER_NAME, s.name)
                    print(" INFO: For material {:s}, shader short name '{:s}' found in IGS_ExpModFile.txt and replaced by long shader name {:s}".format(material.name, s.name, mShaderName))
                    num_slot = 1
                    continue
                except:
                    mShaderName = s.name
                    if not (x[0] for x in SHADERS_LIST_TO_CHECK if mShaderName==x[0]): # Is shader name in SHADERS_LIST_TO_CHECK ?
                        print(" INFO: For material {:s}, shader short name '{:s}' not found in IGS_ExpModFile.txt : shader short name used as the long shader name".format(material.name, s.name))
                    num_slot = 1
                    continue
            else:
                if num_slot == 1:
                    #print(" >>> material.texture_slots {:s} material.name {:s} texture {:s}".format(material.texture_slots, material.name, s.name))
                    mNeverOcclude = 0
                    mSortPriority = 0

                    # FilterMode, AnimateUVs, NumFrames, FPS
                    mFilterMode = 3
                    try:
                        mFilterMode = specific_config.get('Materials:FilterMode', material.name)
                        if not mFilterMode.isdigit():
                            print(" ERROR: For material {:s}, FilterMode {:s} ignored (non numeric value found)".format(material.name, mFilterMode))
                            mFilterMode = 3
                        else:
                            mFilterMode = int(mFilterMode)
                            print(" INFO: FilterMode for material {:s}: {:d}".format(material.name, mFilterMode))
                    except:
                        pass

                    mAnimateUVs = 0
                    try:
                        mAnimateUVs = specific_config.get('Materials:AnimateUVs', material.name)
                        if not mAnimateUVs.isdigit():
                            print(" ERROR: For material {:s}, AnimateUVs {:s} ignored (non numeric value found)".format(material.name, mAnimateUVs))
                            mAnimateUVs = 0
                        else:
                            mAnimateUVs = int(mAnimateUVs)
                            print(" INFO: AnimateUVs for material {:s}: {:d}".format(material.name, mAnimateUVs))
                    except:
                        pass

                    mNumFrames = 0
                    try:
                        mNumFrames = specific_config.get('Materials:NumFrames', material.name)
                        if not mNumFrames.isdigit():
                            print(" ERROR: For material {:s}, NumFrames {:s} ignored (non numeric value found)".format(material.name, mNumFrames))
                            mNumFrames = 0
                        else:
                            mNumFrames = int(mNumFrames)
                            print(" INFO: NumFrames for material {:s}: {:d}".format(material.name, mNumFrames))
                    except:
                        pass

                    mFPS = 0
                    try:
                        mFPS = specific_config.get('Materials:FPS', material.name)
                        if not mFPS.isdigit():
                            print(" ERROR: For material {:s}, FPS {:s} ignored (non numeric value found)".format(material.name, mFPS))
                            mFPS = 0
                        else:
                            mFPS = int(mFPS)
                            print(" INFO: FPS for material {:s}: {:d}".format(material.name, mFPS))
                    except:
                        pass

                    # AlphaTestMode, ZBias, ZBufferMode, VisMod, TwoSided, BackfaceCull, ViewFacing
                    mAlphaTestMode = 0
                    try:
                        mAlphaTestMode = specific_config.get('Materials:AlphaTestMode', material.name)
                        if not mAlphaTestMode.isdigit():
                            print(" ERROR: For material {:s}, AlphaTestMode {:s} ignored (non numeric value found)".format(material.name, mAlphaTestMode))
                            mAlphaTestMode = 0
                        else:
                            mAlphaTestMode = int(mAlphaTestMode)
                            print(" INFO: AlphaTestMode for material {:s}: {:d}".format(material.name, mAlphaTestMode))
                    except:
                        pass

                    mZBias = int(ceil(material.offset_z)) # Round to upper integer
                    try:
                        mZBias = specific_config.get('Materials:ZBias', material.name)
                        if not mZBias.isdigit():
                            print(" ERROR: For material {:s}, ZBias {:s} ignored (non integer value found)".format(material.name, mZBias))
                            mZBias = int(ceil(material.offset_z))
                        else:
                            mZBias = int(mZBias)
                            print(" INFO: ZBias for material {:s}: {:d}".format(material.name, mZBias))
                    except:
                        pass

                    mZBufferMode = 1
                    try:
                        mZBufferMode = specific_config.get('Materials:ZBufferMode', material.name)
                        if not mZBufferMode.isdigit():
                            print(" ERROR: For material {:s}, ZBufferMode {:s} ignored (non integer value found)".format(material.name, mZBufferMode))
                            mZBufferMode = 1
                        else:
                            mZBufferMode = int(mZBufferMode)
                            print(" INFO: ZBufferMode for material {:s}: {:d}".format(material.name, mZBufferMode))
                    except:
                        pass
                        
                    mVisMod = 0
                    try:
                        mVisMod = specific_config.get('Materials:VisMod', material.name)
                        if not mVisMod.isdigit():
                            print(" ERROR: For material {:s}, VisMod {:s} ignored (non numeric value found)".format(material.name, mVisMod))
                            mVisMod = 0
                        else:
                            mVisMod = int(mVisMod)
                            print(" INFO: VisMod for material {:s}: {:d}".format(material.name, mVisMod))
                    except:
                        pass

                    mTwoSided = 0
                    try:
                        mTwoSided = specific_config.get('Materials:TwoSided', material.name)
                        if not mTwoSided.isdigit():
                            print(" ERROR: For material {:s}, TwoSided {:s} ignored (non numeric value found)".format(material.name, mTwoSided))
                            mTwoSided = 0
                        else:
                            mTwoSided = int(mTwoSided)
                            print(" INFO: TwoSided for material {:s}: {:d}".format(material.name, mTwoSided))
                    except:
                        pass

                    mBackfaceCull = 0
                    try:
                        mBackfaceCull = specific_config.get('Materials:BackfaceCull', material.name)
                        if not mBackfaceCull.isdigit():
                            print(" ERROR: For material {:s}, BackfaceCull {:s} ignored (non numeric value found)".format(material.name, mBackfaceCull))
                            mBackfaceCull = 0
                        else:
                            mBackfaceCull = int(mBackfaceCull)
                            print(" INFO: BackfaceCull for material {:s}: {:d}".format(material.name, mBackfaceCull))
                    except:
                        pass
                        
                    mAdditiveAlphaStrength = 128
                    if mAnimateUVs != 0:
                        mAdditiveAlphaStrength = 0
                    
                    mViewFacing = 0
                    if mShaderName.find('ViewFacing') >= 0:
                        if mShaderName.find('Upright') >= 0:
                            mViewFacing = 2
                        else:
                            mViewFacing = 1
                    try:
                        mViewFacing = specific_config.get('Materials:ViewFacing', material.name)
                        if not mViewFacing.isdigit():
                            print(" ERROR: For material {:s}, ViewFacing {:s} ignored (non numeric value found)".format(material.name, mViewFacing))
                            mViewFacing = 0
                        else:
                            mViewFacing = int(mViewFacing)
                            print(" INFO: ViewFacing for material {:s}: {:d}".format(material.name, mViewFacing))
                    except:
                        pass
                   
                    if mAnimateUVs == 0: # Not an animated textures
                        mSurfaceType = b'NO COLLIDE'
                        mUseLightGlow = 0
                        mUsePulseGlow = 0
                        mLightGlowType = b'HEADLAMP'
                        PulseGlowCmTargCol = _igs_rgba_color(0.78,0.78,1.0,1.0)
                        
                        PulseGlowCmMode = 0
                        PulseGlowCmPhase = 0.0
                        PulseGlowCmPeriod = 5
                        
                        PulseGlowCmRptTimeMin = 1.0
                        PulseGlowCmRptTimeMax = 5.0
                        PulseGlowCmExeTimeMin = 1.0
                        PulseGlowCmExeTimeMax = 5.0
                        PulseGlowCmActivationTime = 0.0
                        
                        PulseGlowCmAmplitude = 1
                        
                        PulseGlowCmSeqMode = b'TRAFFIC'
                        PulseGlowCmVertexModify = 0
                        mUnfoggable = 0
                        mForce32Bit = 0

                        mAmbient = _igs_rgba_color(1.0,1.0,1.0,1.0)
                        #mAmbient = _igs_rgba_color(material.ambient,material.ambient,material.ambient,1.0)
                        mEmissive = _igs_rgba_color(material.mirror_color[0],material.mirror_color[1],material.mirror_color[2],1.0)

                    else: # Animated textures
                        mSurfaceType = b''
                        mUseLightGlow = 0
                        mUsePulseGlow = 0
                        mLightGlowType = b''
                        PulseGlowCmTargCol = _igs_rgba_color(0.0,0.0,0.0,0.0)
                        
                        PulseGlowCmMode = 0
                        PulseGlowCmPhase = 0.0
                        PulseGlowCmPeriod = 0
                        
                        PulseGlowCmRptTimeMin = 0.0
                        PulseGlowCmRptTimeMax = 0.0
                        PulseGlowCmExeTimeMin = 0.0
                        PulseGlowCmExeTimeMax = 0.0
                        PulseGlowCmActivationTime = 0.0
                        
                        PulseGlowCmAmplitude = 0
                        
                        PulseGlowCmSeqMode = b''
                        PulseGlowCmVertexModify = 0
                        mUnfoggable = 0
                        mForce32Bit = 0
                        mAmbient = _igs_rgba_color(0.0,0.0,0.0,0.0)
                        mEmissive = _igs_rgba_color(material.mirror_color[0],material.mirror_color[1],material.mirror_color[2],0.0)

                    # Custom Ambient?
                    try:
                        mAmbient_value = specific_config.get('Materials:Ambient', material.name)
                        if not mAmbient_value.replace('.','',1).isdigit():
                            print(" ERROR: For material {:s}, Ambient {:s} ignored (non numeric value found)".format(material.name, mAmbient))
                            mAmbient = _igs_rgba_color(1.0,1.0,1.0,1.0)
                        else:
                            mAmbient = _igs_rgba_color(float(mAmbient_value), float(mAmbient_value), float(mAmbient_value),1.0)
                            print(" INFO: Ambient for material {:s}: {:s} (for red, green and blue)".format(material.name, mAmbient_value))
                    except:
                        pass
                    
                    mDiffuse = _igs_rgba_color(material.diffuse_color[0],material.diffuse_color[1],material.diffuse_color[2],1.0)
                    mSpecular = _igs_rgba_color(material.specular_color[0],material.specular_color[1],material.specular_color[2],1.0)
                    mEmissiveStrength = material.emit
                    
                    mSpecularPower = material.specular_intensity * material.specular_hardness
                    mLMCastShadows = 1
                    mLMKeepVertexColours = 0
                    if mAnimateUVs == 0: # Not an animated textures
                        mLMGenerateShadows = 1
                        mLMTexelsPerMetre = 4
                    else: # Animated textures
                        mLMGenerateShadows = 0
                        mLMTexelsPerMetre = 0
                    mComment= ''
                    mExtraData = _IGDataBlockContainer(0, 0, b'', [0, 0])
                    num_slot += 1
                    mNumRenderStages = 0

                    IGRenderSmTextureName = []
                    IGRenderSmPadding = []
                    IGRenderSmUVChannelIndex = []
                    IGRenderSmFilterMode = []
                    IGRenderSmUVOperation = []
                    IGRenderSmUVWrapU = []
                    IGRenderSmUVWrapV = []
                    IGRenderSmMipLODBias = []
                    IGRenderSmAnimateUVs = []
                    IGRenderSmNumFrames = []
                    IGRenderSmFPS = []
                    IGRenderSmScrollUVs = []
                    IGRenderSmScrollU = []
                    IGRenderSmScrollV = []
                    IGRenderSmArguments = []

                else:
                    mNumRenderStages += 1
                    num_slot += 1
                    #print(" >>> num slot {:d} mNumRenderStages {:d}".format(num_slot, mNumRenderStages))

                # Material data for the slot
                #print(" >>> mNumRenderStages {:d} / num_slot {:d} ".format(mNumRenderStages, num_slot))
                try:
                    currentTextureFullName = s.texture.image.filepath
                except:
                    print(" ERROR: Missing image file in material {:s}".format(material.name))
                    return None, (num_slot - 1), add_face_tag_description
                currentTextureName = bpy.path.basename(currentTextureFullName)
                IGRenderSmTextureName.append(add_texture_to_list(target_textures_directory, currentTextureName, textures_list, specific_config))

                IGRenderSmPadding.append(b'')
                IGRenderSmUVChannelIndex.append(mNumRenderStages)   # MAJ_1_4_0 /  Channel index
                IGRenderSmFilterMode.append(mFilterMode)
                IGRenderSmUVOperation.append(0)
                IGRenderSmUVWrapU.append(0)
                IGRenderSmUVWrapV.append(0)
                if mNumRenderStages == 0:
                    if mShaderName.find('Spec') >= 0 and mShaderName.find('BumpSpecEnv') == 0:
                        IGRenderSmMipLODBias.append(-1.0)
                    else:
                        if mShaderName.find('FacingFlora') >= 0:
                            #IGRenderSmMipLODBias.append(0.0)
                            IGRenderSmMipLODBias.append(-2.0) # 1.4.5
                            #add_face_tag_description = True # (Code added in 1.4.3 to deal with the FaceTagDescription but not actually needed (also found for some other shaders)
                        else:
                            mMipLODBias = -2.0

                            try:
                                mMipLODBias = specific_config.get('Materials:MipLODBias', material.name)
                                if not isfloat(mMipLODBias):
                                    print(" ERROR: For material {:s}, MipLODBias {:s} ignored (non float value found)".format(material.name, mMipLODBias))
                                    mMipLODBias = -2.0
                                else:
                                    mMipLODBias = float(mMipLODBias)
                                    print(" INFO: MipLODBias for material {:s}: {:f}".format(material.name, mMipLODBias))
                            except:
                                pass
                            IGRenderSmMipLODBias.append(mMipLODBias)

                    try:
                        UVArguments = specific_config.get('UVArguments', material.name)
                        UVArguments = str(UVArguments).replace(' ','') # remove spaces
                        s = UVArguments.split(',')
                        if len(s) != 6:
                            print(" ERROR: UVArguments for material {:s} ignored (6 values expected - {:d} found)".format(material.name, len(s)))
                            IGRenderSmArguments.append((0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
                        elif not s[0].replace('.','',1).isdigit() or not s[1].replace('.','',1).isdigit() or not s[2].replace('.','',1).isdigit() or \
                        not s[3].replace('.','',1).isdigit() or not s[4].replace('.','',1).isdigit() or not s[5].replace('.','',1).isdigit():
                            print(" ERROR: UVArguments for material {:s}: {:s} {:s} {:s} {:s} {:s} {:s} ignored (non numeric value found)".format(material.name, s[0], s[1], s[2], s[3], s[4], s[5]))
                            IGRenderSmArguments.append((0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
                        else:
                            print(" INFO: UVArguments for material {:s}: {:s} {:s} {:s} {:s} {:s} {:s}".format(material.name, s[0], s[1], s[2], s[3], s[4], s[5]))
                            IGRenderSmArguments.append((float(s[0]), float(s[1]), float(s[2]), float(s[3]), float(s[4]), float(s[5])))
                    except:
                        IGRenderSmArguments.append((0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
                        pass

                elif mNumRenderStages == 1:
                    if mShaderName.find('BumpSpecEnv') >= 0:
                        IGRenderSmMipLODBias.append(-1.0)
                    elif mShaderName.find('Spec') >= 0:
                        IGRenderSmMipLODBias.append(-0.0)
                    else:
                        IGRenderSmMipLODBias.append(0.0)
                elif mNumRenderStages == 2 or (mShaderName.find('WeatherEffects') >= 0 and mNumRenderStages == 3):
                    IGRenderSmMipLODBias.append(0.0)
                if mNumRenderStages != 0:
                    IGRenderSmArguments.append((0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
                IGRenderSmAnimateUVs.append(mAnimateUVs)
                IGRenderSmNumFrames.append(mNumFrames)
                if mFPS == 0:
                    # Use model setting
                    IGRenderSmFPS.append(int(bpy.context.scene.render.fps / bpy.context.scene.render.fps_base))
                else:
                    # FPS was set in configuration file
                    IGRenderSmFPS.append(mFPS)
                IGRenderSmScrollUVs.append(0)
                IGRenderSmScrollU.append(0.0)
                IGRenderSmScrollV.append(0.0)

                if mShaderName.find('WeatherEffects') >= 0 and num_slot == 2:
                    name_split = material.name.split('_')
                    if len(name_split) != 2 or name_split[0] != 'weatherglass' or not name_split[1].isdigit():
                        print(" ERROR: With shader name TrainGlassWeatherEffects.fx, material name must be weatherglass_1, weatherglass_2, etc.. ('{:s}' was found)".format(material.name))
                        return None, (num_slot - 1), add_face_tag_description

                    mNumRenderStages += 1
                    num_slot += 1
                    
                    IGRenderSmTextureName.append(add_texture_to_list(target_textures_directory, currentTextureName, textures_list, specific_config))
                    
                    IGRenderSmPadding.append(b'')
                    IGRenderSmUVChannelIndex.append(0)
                    IGRenderSmFilterMode.append(3)
                    IGRenderSmUVOperation.append(0)
                    IGRenderSmUVWrapU.append(0)
                    IGRenderSmUVWrapV.append(0)
                    IGRenderSmMipLODBias.append(-1.0)
                    IGRenderSmAnimateUVs.append(0)
                    IGRenderSmNumFrames.append(0)
                    IGRenderSmFPS.append(int(bpy.context.scene.render.fps / bpy.context.scene.render.fps_base))
                    IGRenderSmScrollUVs.append(0)
                    IGRenderSmScrollU.append(0.0)
                    IGRenderSmScrollV.append(0.0)
                    # Always 0.0 ,...
                    IGRenderSmArguments.append((0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
                    #print(" >>> mNumRenderStages 2 {:d}".format(mNumRenderStages))
                    
                material_found = True
            
    if material_found:
        for i, sublist in enumerate(SHADERS_LIST_TO_CHECK):
            if sublist[0] == mShaderName:
                #print(" >>>> Shader {:s} / SHADERS_LIST_TO_CHECK[i][1] {:d} / (num_slot - 1) {:d}".format(sublist, SHADERS_LIST_TO_CHECK[i][1], (num_slot - 1)))
                if SHADERS_LIST_TO_CHECK[i][1] != (num_slot - 1):
                    print(" ERROR: Missing slots for shader name {:s} used in material named {:s}, {:d} texture slots must be set in Blender (+ the first slot with the shader name)".format(mShaderName, material.name, SHADERS_LIST_TO_CHECK[i][2]))
                    return None, (num_slot - 1), add_face_tag_description

        #print(" >>> IGRenderSmTextureName {:d}".format(len(IGRenderSmTextureName))
        new_material = _IGMaterials(material.name, mShaderName, mNumRenderStages + 1, \
        IGRenderSmTextureName, IGRenderSmPadding, IGRenderSmUVChannelIndex, IGRenderSmFilterMode, IGRenderSmUVOperation, IGRenderSmUVWrapU, \
        IGRenderSmUVWrapV, IGRenderSmMipLODBias, \
        IGRenderSmAnimateUVs, IGRenderSmNumFrames, IGRenderSmFPS, IGRenderSmScrollUVs, IGRenderSmScrollU, IGRenderSmScrollV, IGRenderSmArguments, \
        mNeverOcclude, mSortPriority, mZBias, mAlphaTestMode, mAdditiveAlphaStrength, mZBufferMode, \
        mBackfaceCull, mTwoSided, mViewFacing, mVisMod, mSurfaceType, mUseLightGlow, mUsePulseGlow, mLightGlowType, \
        PulseGlowCmTargCol, PulseGlowCmMode, PulseGlowCmPhase, PulseGlowCmPeriod, PulseGlowCmAmplitude, PulseGlowCmRptTimeMin, \
        PulseGlowCmRptTimeMax, PulseGlowCmExeTimeMin, PulseGlowCmExeTimeMax, PulseGlowCmActivationTime, PulseGlowCmSeqMode, PulseGlowCmVertexModify, \
        mUnfoggable, mForce32Bit, mAmbient, mDiffuse, mEmissive, \
        mEmissiveStrength, mSpecular, mSpecularPower, mLMCastShadows, mLMGenerateShadows, mLMKeepVertexColours, mLMTexelsPerMetre, mComment, mExtraData)

        return new_material, (num_slot - 1), add_face_tag_description
    else:
        return None, (num_slot - 1), add_face_tag_description


def get_autosmooth_normal(mesh, fop, mesh_vi):
# Function adapted and fixed from the trainz exporter (thanks!)
    '''check if smoothing has to be applied and return the depending normal'''
    ##  is at least one neighbourface smooth, the we return the vertex
    ##  normal(smooth applied), else the face normal will be returned
    result = fop.normal  # init with the normal for the un-smooth case

    ##  faulty (none-planar) faces may have a zero-length normal, and without
    ##  direction you can't calculate direction difference
    if fop.normal.length > 0.0:
        for p in mesh.tessfaces:
            if ( p != fop ) and ( mesh_vi in p.vertices ) and ( p.normal.length > 0.0):
                angle = fop.normal.angle(p.normal)
                if  angle <= mesh.auto_smooth_angle:
                    result = mesh.vertices[mesh_vi].normal
    return result

    
def print_config(config, config_name):
    print('\n Configuration file: ', config_name)
    for section in config.sections():
        print('------ ',section,' ------')
        for option in config.items(section):
            print(option)
    print('\n')
      

def getLOD(Name):
# Get a LOD compliant name: like 1_0100_name
# A LOD compliant name is a digit (level), an underscore, 4 digits (visible distance), and underscore and the name, and the total length is <= 31 (24 for the name)
    MAX_NAME_WITH_LOD = 31
    MAX_NAME_WITHOUT_LOD = 24
    SIZE_DISTANCE_DIGITS = 4

    LODlevel = '1'
    LODdist = '1000'
    name_split = Name.split('_')
    if len(name_split) >= 2:
        if name_split[0].isdigit() and name_split[1].isdigit():
            if len(name_split[0]) == 1 and len(name_split[1]) == SIZE_DISTANCE_DIGITS:
                LODlevel = name_split[0]
                LODdist = name_split[1]
                if len(Name) > MAX_NAME_WITH_LOD:
                    NewName = Name[:MAX_NAME_WITH_LOD]
                    print(" WARNING: Object '{:s}' processed using a truncated name at {:d} characters: '{:s}' (to comply with LOD naming rule). The converted name is used for LOD processing and may be used if the object is head of group.".format(Name, MAX_NAME_WITH_LOD, NewName))
                    return LODlevel, LODdist, NewName
                else:
                    return LODlevel, LODdist, Name
            else:
                if len(name_split[0]) == 1: # LODlevel
                    if len(name_split[1]) >= SIZE_DISTANCE_DIGITS: # LODdist
                    # Truncate
                        LODlevel = name_split[0]
                        LODdist = name_split[1]
                        NewName = "_".join([LODlevel, LODdist[:SIZE_DISTANCE_DIGITS], name_split[2]])
                        print(" WARNING: Object '{:s}' processed as '{:s}' (to comply with LOD naming rule). The converted name is used for LOD processing and may be used if the object is head of group.".format(Name, NewName[:MAX_NAME_WITH_LOD]))
                        return LODlevel, LODdist[:SIZE_DISTANCE_DIGITS], NewName[:MAX_NAME_WITH_LOD]
                    else:
                    # Pad
                        LODlevel = name_split[0]
                        LODdist = name_split[1]
                        NewName = "_".join([LODlevel, LODdist.ljust(SIZE_DISTANCE_DIGITS,'0'), name_split[2]])
                        print(" WARNING: Object '{:s}' processed as '{:s}' (to comply with LOD naming rule). The converted name is used for LOD processing and may be used if the object is head of group.".format(Name, NewName[:MAX_NAME_WITH_LOD]))
                        return LODlevel, LODdist.ljust(SIZE_DISTANCE_DIGITS,'0'), NewName[:MAX_NAME_WITH_LOD]
                else:
                    NewName = "_".join([LODlevel, LODdist, Name[:MAX_NAME_WITHOUT_LOD]])
                    print(" WARNING: Object '{:s}' processed as '{:s}' (to comply with LOD naming rule). The converted name is used for LOD processing and may be used if the object is head of group.".format(Name, NewName))
                    return LODlevel, LODdist, NewName
        else:
            NewName = "_".join([LODlevel, LODdist, Name[:MAX_NAME_WITHOUT_LOD]])
            print(" WARNING: Object '{:s}' processed as '{:s}' (to comply with LOD naming rule). The converted name is used for LOD processing and may be used if the object is head of group.".format(Name, NewName))
            return LODlevel, LODdist, NewName #(LODlevel + '_' + LODdist + '_' + Name)
    else:
        NewName = "_".join([LODlevel, LODdist, Name[:MAX_NAME_WITHOUT_LOD]])
        print(" WARNING: Object '{:s}' processed as '{:s}' (to comply with LOD naming rule). The converted name is used for LOD processing and may be used if the object is head of group.".format(Name, NewName))
        return LODlevel, LODdist, NewName # (LODlevel + '_' + LODdist + '_' + Name)


def update_bounding_box(Vmin_bounding_box, Vmax_bounding_box, new_vector):

    minx = Vmin_bounding_box.x
    miny = Vmin_bounding_box.y
    minz = Vmin_bounding_box.z
    maxx = Vmax_bounding_box.x
    maxy = Vmax_bounding_box.y
    maxz = Vmax_bounding_box.z

    #print(" >>> Bounding box update : new_vector {:f} {:f} {:f}".format(new_vector.x, new_vector.y, new_vector.z))
    if new_vector.x < minx:
        minx = new_vector.x
    elif new_vector.x > maxx:
        maxx = new_vector.x

    if new_vector.y < miny:
        miny = new_vector.y
    elif new_vector.y > maxy:
        maxy = new_vector.y

    if new_vector.z < minz:
        minz = new_vector.z
    elif new_vector.z > maxz:
        maxz = new_vector.z

    Vmin_bounding_box = Vector([minx, miny, minz])
    Vmax_bounding_box = Vector([maxx, maxy, maxz])
    #print(" >>> Bounding box update 2 : minx {:f}, miny {:f}, minz {:f} // maxx {:f}, maxy {:f}, maxz {:f}".format(\
    #(Vmin_bounding_box.x, Vmin_bounding_box.y, Vmin_bounding_box.z, Vmax_bounding_box.x, Vmax_bounding_box.y, Vmax_bounding_box.z))
    return Vmin_bounding_box, Vmax_bounding_box


def same_vertex(vertex_coord, vertex_normal, nb_uvMaps, vertex_uv, vertex_colour, vertex_material_offset, vertex_obj):
#    return False

    # check positional absolute
    if (vertex_coord.x != vertex_obj.mPoint.x):
        return False
    if (vertex_coord.y != vertex_obj.mPoint.y):
        return False
    if (vertex_coord.z != vertex_obj.mPoint.z):
        return False
    if (nb_uvMaps != vertex_obj.mNumValidUVs):
        return False

    if (vertex_material_offset != vertex_obj.mMaterial):
        return False

    # weld normal
    if (abs(vertex_obj.mNormal.x - vertex_normal.x) >= gVertexWeldNormalThreshold):
        return False
    if (abs(vertex_obj.mNormal.y - vertex_normal.y) >= gVertexWeldNormalThreshold):
        return False
    if (abs(vertex_obj.mNormal.z - vertex_normal.z) >= gVertexWeldNormalThreshold):
        return False

    # weld colour
    if (abs(vertex_obj.mColour.r - vertex_colour.r) >= gVertexWeldColourThreshold):
        return False
    if (abs(vertex_obj.mColour.g - vertex_colour.g) >= gVertexWeldColourThreshold):
        return False
    if (abs(vertex_obj.mColour.b - vertex_colour.b) >= gVertexWeldColourThreshold):
        return False
    if (abs(vertex_obj.mColour.a - vertex_colour.a) >= gVertexWeldColourThreshold):
        return False

    # check UVs for all uv maps
    for NumUV in range(vertex_obj.mNumValidUVs):  # MAJ_1_4_0
        if (abs(vertex_obj.mUV[NumUV].u - vertex_uv[NumUV].u) >= gVertexWeldUVThreshold):
            return False
        if (abs(vertex_obj.mUV[NumUV].v -vertex_uv[NumUV].v) >= gVertexWeldUVThreshold):
            return False
    
    return True

    
def find_vertex(vertex_coord, vertex_normal, nb_uvMaps, vertex_uv, vertex_colour, vertex_material_offset, object_vertices_list, start_of_list):

    for vertex_index, vertex_obj in enumerate(object_vertices_list[start_of_list:]):
        if same_vertex(vertex_coord, vertex_normal, nb_uvMaps, vertex_uv, vertex_colour, vertex_material_offset, vertex_obj):
            return vertex_obj, vertex_index + start_of_list
        
    return None, -1
    
def get_UVMaps(mesh, uv_name): # MAJ_1_4_0 (new)
    # Get the uv map (Object data / uvmap section)
    #print(" >>> get_UVMaps uv_name", uv_name)
    uv_layer = None
    if uv_name == "":
        uv_layer = mesh.tessface_uv_textures[0]
    else:
        i = 0
        while uv_layer == None and i<len(mesh.tessface_uv_textures):
            if mesh.tessface_uv_textures[i].name == uv_name:
                uv_layer = mesh.tessface_uv_textures[i]
                #print(" >>> get_UVMaps i", i)
                return uv_layer
            i += 1
    return uv_layer
    
def process_mesh_geometry(bpy, group_wrapper, obj, obj_matrix, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, applied_translation, Vmin_bounding_box, Vmax_bounding_box, materials_list, textures_list, object_triangles_list, \
object_vertices_list, vertices_count, vertices_count_start, triangles_count, materials_count, default_config, specific_config, mExtraData, smaller_IGS_file, object_vertices_params_list):

    try:
        data = obj.to_mesh(bpy.context.scene, True, 'PREVIEW') # Create a Mesh datablock with modifiers applied (Applying modifiers 'preview' settings)
        #print(" >>> preview", obj.name)
    except:
        data = None

    add_face_tag_description = False
    triangles_count_current = triangles_count
    if data or (obj.type == 'EMPTY' and not first_obj_found): # Object with mesh data or empty at start of group
        # http://blenderscripting.blogspot.fr/search?updated-max=2012-03-20T03:56:00-07:00&max-results=7
        obj_matrix = obj.matrix_world.copy()
        #print(' >>> ', obj.matrix_world)
        if not first_obj_found:
            # Change of coordinates system
            #print(' >>>>>>> 1st matrix', obj_matrix)
            first_obj_transform = Matrix.Rotation(radians(90), 4, 'X')
            first_obj_transform *= Matrix.Scale(-1, 4, Vector((0, 0, 1)))
            first_obj_matrix = obj.matrix_world.copy()
            first_obj_matrix.translation = first_obj_matrix.to_translation() - applied_translation
            if obj.type == 'EMPTY':
                first_obj_found = True
            #print(' >>> first_obj_matrix ', first_obj_matrix, first_obj_transform)

        #print(" >>> data found {:s} materials_count={:d} vertices_count_start={:d} first_obj_found {:d}"% (obj.name, materials_count, vertices_count_start, first_obj_found))

        if data:
            data.calc_tessface() # Compute and update the list of tessellated faces (faces of 3 or 4 vertices)

            if not data.tessface_uv_textures:
                # to_mesh() created a new mesh so remove it
                bpy.data.meshes.remove(data)     
                print(" ERROR: Blender Mesh object    '{:s}' has no collection of UV maps for tessellated faces (check material assignement).".format(obj.name))
                return END_ERROR, vertices_count, triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description
            
            mat_ls = data.materials
            if len(mat_ls) == 0:
                # to_mesh() created a new mesh so remove it
                bpy.data.meshes.remove(data)     
                print(" ERROR: Blender Mesh object    '{:s}' has no material assigned.".format(obj.name))
                return END_ERROR, vertices_count, triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description

            for mat in data.materials:
                if mat is not None: # skip empty material slots
                    mat_index = -1
                    #print(" >>> Look for material {:s} mat {:s}".format(obj.name, mat.name))
                    for i, sublist in enumerate(materials_list):
                        if sublist[0] == mat.name:
                            mat_index = i
                            #print(" >>>> Mat found {:s} ".format(sublist))
                            break
                    if mat_index == -1:

                        material_IGMaterials, nb_slots, add_face_tag_description_mat = add_igs_material(bpy, mat, default_config, specific_config, textures_list)
                        if material_IGMaterials is None:
                            print(" ERROR: No valid texture found in material '{:s}' in Blender Mesh object    '{:s}' (check shader name and subsequent slots).".format(mat.name, obj.name))
                            return END_ERROR, vertices_count, triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description
                        mat_index = len(materials_list)
                        materials_count += 1
                        if add_face_tag_description_mat:
                            add_face_tag_description = True

                        materials_list.append([mat.name, material_IGMaterials, nb_slots])
                        material_found = True
                    #else:
                        #print(" >>> existing mat_index {:d}".format(mat_index))
                        #material_IGMaterials = materials_list[mat_index][1]
                        #nb_slots = materials_list[mat_index][2]
                        
                    #print(" >>> Object material loop {:s} mat {:s} nb_slots {:d} materials_count {:d}".format(obj.name, mat.name, nb_slots, materials_count))

            # Get vertices, normals and UVs from the faces
            if first_obj_found:
                first_vertex_processed = True
            else:
                first_vertex_processed = False
            
            for face_num, face in enumerate(data.tessfaces):
                f_v = face.vertices
                f_mat = data.materials[face.material_index]
                if f_mat is None:
                    print(" ERROR: Blender Mesh object    '{:s}' has a face with no texture assigned.".format(obj.name))
                    return END_ERROR, vertices_count, triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description
                for i, sublist in enumerate(materials_list):
                    if sublist[0] == f_mat.name:
                        f_mat_nb_slots = sublist[2]
                        face_mat_index = i
                        #print(" >>>> f_mat_nb_slots {:d} f_mat.name {:s} face.material_index {:d} ".format(f_mat_nb_slots, f_mat.name, face.material_index))
                        break

                mat_igs_offset = IGS_HEADER_SIZE + face_mat_index * IGS_MATERIAL_SIZE

                # In a given mesh, a slot may have a specific UV mapping (MAJ_1_4_0)
                f_uv_list = []
                nb_uvMaps = 0
                for numUV in range(f_mat_nb_slots): # MAJ_1_4_0
                    #print(" >>>> numUV / f_mat_nb_slots ", numUV, f_mat_nb_slots)
                    uv_name = ""
                    if bpy.data.materials[f_mat.name].texture_slots[numUV+1] is not None:
                        uv_name = bpy.data.materials[f_mat.name].texture_slots[numUV+1].uv_layer
                    uv_layer = get_UVMaps(data, uv_name)
                    if uv_layer is not None:
                        nb_uvMaps += 1
                        f_uv_list.append(uv_layer.data[face_num].uv)
                        #print(" >>>> f_mat.name {:s} / nb_uvMaps {:d}".format(f_mat.name, nb_uvMaps))

                # Triangle or quad
                # ----------------

                new_colour = _igs_rgba_color(0.0, 0.0, 0.0, 0.0)
                #print(" >>>> numUV {:d} / nb_uvMaps {:d}".format(numUV, nb_uvMaps))
                for nb in range(2):
                    triangle_vertices_list = []
                    triangle_vertex_indices_list = []
                    if nb == 0:
                        # vertex_indices_list = (f_v[0], f_v[1], f_v[2])
                        # facesUV_list = (f_uv[0], f_uv[1], f_uv[2]) # UV coordinates
                        # normals_list = (data.vertices[f_v[0]].normal, data.vertices[f_v[1]].normal, data.vertices[f_v[2]].normal)
                        # points_coord_list = (data.vertices[f_v[0]].co, data.vertices[f_v[1]].co, data.vertices[f_v[2]].co)
                        num1, num2, num3 = 0, 1, 2
                    else:
                        if len(f_v) == 4:  # Quad: add another triangle
                            num1, num2, num3 = 0, 2, 3
                        else:
                            break
                    for vert_num in (num1, num2, num3):
                        # convert local object coordinates -> world - > local coordinates of first object -> axis change
                        local_vertex = obj_matrix * data.vertices[f_v[vert_num]].co - applied_translation
                        local_vertex = first_obj_matrix.inverted() * local_vertex
                        local_vertex = first_obj_transform * local_vertex
                        if not first_vertex_processed:
                            first_vertex_processed = True
                            # Initialize bounding box in world coordinates
                            Vmin_bounding_box = local_vertex
                            Vmax_bounding_box = local_vertex
                        else:
                            # Update bounding box in world coordinates
                            Vmin_bounding_box, Vmax_bounding_box = update_bounding_box(Vmin_bounding_box, Vmax_bounding_box, local_vertex)
                        # Check if vertex already exists and get normal according to smooth settings
                        if face.use_smooth: # Smooth button was pressed in the Tool Shelf
                        # http://wiki.blender.org/index.php/Doc:2.6/Manual/Modeling/Meshes/Smoothing
                            if data.use_auto_smooth:
                                #  Use vertex normal or face normal according to auto_smooth angle
                                new_normal_rot = obj_matrix.to_quaternion() * get_autosmooth_normal(data, face, f_v[vert_num])
                            else:
                                new_normal_rot = obj_matrix.to_quaternion() * data.vertices[f_v[vert_num]].normal # Smooth: use vertex normal
                        else:
                            new_normal_rot = obj_matrix.to_quaternion() * face.normal # No smooth: use face normal
                        new_normal_rot = first_obj_matrix.to_quaternion().inverted() * new_normal_rot
                        new_normal_rot = first_obj_transform * new_normal_rot
                        new_UV0 = []
                        for numUV in range(nb_uvMaps): # MAJ_1_4_0
                            f_uv = f_uv_list[numUV]
                            new_UV0.append(_igs_UV(f_uv[vert_num][0], 1 - f_uv[vert_num][1]))
                        if smaller_IGS_file:
                            vertex_index = -1
                            #new_vertex, vertex_index = find_vertex(local_vertex, new_normal_rot, nb_uvMaps, new_UV0, new_colour, mat_igs_offset, object_vertices_list, 0) # MAJ
                            try:  # MAJ2
                                vertex_index = object_vertices_params_list.index((local_vertex.x, local_vertex.y, local_vertex.z, nb_uvMaps, mat_igs_offset))
                                # Vertex found: compare other parameters
                                new_vertex, vertex_index = find_vertex(local_vertex, new_normal_rot, nb_uvMaps, new_UV0, new_colour, mat_igs_offset, object_vertices_list, vertex_index) # MAJ
                            except:
                                new_vertex = None
                                vertex_index = -1
                                
                        else:
                            new_vertex = None

                        if new_vertex:
                            triangle_vertex_indices_list.append(vertex_index)
                            #print(' >>>??? vertex_index ', vertex_index)
                        else:  # None
                            new_point = _igs_vector4(local_vertex.x, local_vertex.y, local_vertex.z, 1.0)
                            new_normal = _igs_vector4(new_normal_rot.x, new_normal_rot.y, new_normal_rot.z, 1.0) # Smoothed (vertex normal used)

                            new_vertex = _IGVertex(new_point, new_normal, nb_uvMaps, new_UV0, new_colour, mat_igs_offset, 0, [], 0 , mExtraData)
                            object_vertices_list.append(new_vertex)
                            if smaller_IGS_file:
                                object_vertices_params_list.append((local_vertex.x, local_vertex.y, local_vertex.z, nb_uvMaps, mat_igs_offset)) # MAJ2
                            triangle_vertex_indices_list.append(vertices_count)
                            #print(' >>>??? triangle_vertex_indices_list append nb_uvMaps / vertices_count ', nb_uvMaps, vertices_count)
                            vertices_count += 1
                        triangle_vertices_list.append(new_vertex)
                    # mVertex, mNormal, mHiddenEdges, mMaterial, mSMGroup, mTagId, mExtraData, vertices_list
                    # Fill mVertex with vertex index before future update to offset value for the final igs file.
                    face_normal_rot = obj_matrix.to_quaternion() * face.normal
                    face_normal_rot = first_obj_matrix.to_quaternion().inverted() * face_normal_rot
                    face_normal_rot = first_obj_transform * face_normal_rot
                    face_normal = _igs_vector4(face_normal_rot.x, face_normal_rot.y, face_normal_rot.z, 1.0)
                    new_triangle = _IGTriangle(triangle_vertex_indices_list, face_normal, 0, mat_igs_offset, 0, 0, mExtraData, vertices_count_start) # MAJ
                    object_triangles_list.append(new_triangle)
                    triangles_count += 1
                    #print(' >>>??? process_mesh_geometry triangles_count, vertices_count_start ', triangles_count, vertices_count_start)
            
            # to_mesh() created a new mesh so remove it
            bpy.data.meshes.remove(data)
            first_obj_found = True
        
    if triangles_count_current == triangles_count:
        print(" ERROR: Blender Mesh object    '{:s}' has no triangles.".format(obj.name))
        return END_ERROR, vertices_count, triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description

    return END_OK, vertices_count, triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, first_obj_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description


def create_mesh_igs(grp, matrix, current_id, vertices_count, triangles_count, triangles_count_start, vertices_count_start, \
Vmin_bounding_box, Vmax_bounding_box, mExtraData):

    children_count = len(grp.child_groups_list)
    if children_count > IGF_MAX_CHILDREN:
        print(" ERROR: more than {:d} children for object {:s} (check groups list above)".format(IGF_MAX_CHILDREN, grp.name))
        return None, None, None
    
    if vertices_count == 0: # Empty alone in a group (should not occur as empty's are dropped when scanning the model to build groups)
        new_IGfMesh, new_geomLOD = None, None
        mNumLODs = 0
        mGeometry = 0
    else: # Mesh
        MyPointMin = _igs_vector4(Vmin_bounding_box.x, Vmin_bounding_box.y, Vmin_bounding_box.z, 1.0)
        MyPointMax = _igs_vector4(Vmax_bounding_box.x, Vmax_bounding_box.y, Vmax_bounding_box.z, 1.0)
        bounding_box = _igs_bounding_box(MyPointMin, MyPointMax)
        
        # mIsSkinned, mNumVertices, mVertices, mNumTriangles, mTriangles, mBoundingBox, mExtraData
        new_IGfMesh = _IGGeom(0, vertices_count, vertices_count_start, triangles_count, triangles_count_start, bounding_box, mExtraData)
        
        # VisibleDistance, pGeometry, mExtraData
        new_geomLOD = _IGGeometryLOD(grp.LODdist, current_id, mExtraData)
        
        mNumLODs = 1
        mGeometry = current_id

    myMatrix = _igs_Matrix4x4(matrix[0][0], matrix[2][0], matrix[1][0], matrix[3][0], \
    matrix[0][2], matrix[2][2], matrix[1][2], matrix[3][2], \
    matrix[0][1], matrix[2][1], matrix[1][1], matrix[3][1], \
    matrix[0][3], matrix[2][3], matrix[1][3], matrix[3][3])
    myMatrix2 = _igs_Matrix4x4(matrix[0][0], matrix[2][0], matrix[1][0], matrix[3][0], \
    matrix[0][2], matrix[2][2], matrix[1][2], matrix[3][2], \
    matrix[0][1], matrix[2][1], matrix[1][1], matrix[3][1], \
    matrix[0][3], matrix[2][3], matrix[1][3], matrix[3][3])

    # mName, mNameUID, mNumLODs, mGeometry, mTM, mParent, mNumChildren, mChildren, ExtraData, parent_group, child_groups_list
    new_instance = _IGInstance(grp.name, current_id, mNumLODs, mGeometry, myMatrix, 0, children_count, [], \
    mExtraData, myMatrix2, grp.parent_group, grp.child_groups_list)
    grp.IGinstance = new_instance
    
    return new_IGfMesh, new_geomLOD, new_instance


class group_wrapper(object):
    '''Class to group objects along criterias in order to join them into a single object'''

    __slots__ = "LODlevel", "LODdist", "keyword", "parent_group", "name", "objects_list", "child_groups_list", "IGinstance_id", "IGinstance"

    def __init__(self, LODlevel, LODdist, keyword, parent_group, name):
        self.LODlevel = LODlevel
        self.LODdist = LODdist
        self.keyword = keyword
        self.parent_group = parent_group
        self.name = name
        self.objects_list = []
        self.child_groups_list = []
        self.IGinstance_id = 0 # Id of the related IGinstance
        self.IGinstance = None # Related IGinstance

    def short_dump(self):
        print("--- group_wrapper --- {:s} LODlevel {:s} / LODdist {:s} / keyword {:s} ({:d} items)".format(self.name, self.LODlevel, self.LODdist, self.keyword,  len(self.objects_list)))

    def dump(self):
        if len(self.objects_list) == 1:
            print("--- Group --- {:s} (1 item)".format(self.name))
        else:
            print("--- Group --- {:s} ({:d} items)".format(self.name, len(self.objects_list)))
        # print("    LOD level {:s} / LOD distance {:s} / keyword {:s}".format(self.LODlevel, self.LODdist, self.keyword))
        print("    LOD level {} / LOD distance {} / keyword {}".format(self.LODlevel, self.LODdist, self.keyword))
        if self.parent_group is None:
            print("    Parent group: None")
        else:
            print("    Parent group: {:s}".format(self.parent_group.name))
        print("    Group objects:")
        for obj in self.objects_list:
            print("                  {:s}".format(obj.name))
        if len(self.child_groups_list) != 0:
            print("    Group children:")
        for obj in self.child_groups_list:
            print("                     {:s}".format(obj.name))
        print("")

def group_wrapper_dump(groups_list):        
    print(" --- {:d} groups ---".format(len(groups_list)))
    print("")
    for grp in groups_list:
        grp.dump()
            
def same_group(hierarchy_processing, current_group, new_LODlevel, new_LODdist, LODdist_parent, new_keyword, keyword_type, obj_parent):
    if len(current_group.objects_list) == 0:
        return False
    if current_group.LODlevel != new_LODlevel or current_group.LODdist != new_LODdist:
        return False

    # LODlevel == LODlevel_parent and LODdist != LODdist_parent and parent_object_keyword != None (example: 1_0100_Parent wwith 2 children 1_0030_Child1 and |_1_0030_Child2, then 1_0030_Child* are merged)
    if (int(new_LODlevel) > 1) or (new_LODdist != LODdist_parent): # MAJ_1_4_1 / 1.4.4
        if obj_parent != current_group.objects_list[0].parent:
            # They must be from the same parent
            return False
        #else:
        #    print(" >> (obj_parent == current_group.objects_list[0].parent) ", obj_parent)

    if current_group.keyword != new_keyword:
        return False

    if keyword_type == KEYWORD_CUSTOM:
        return True
    
    if hierarchy_processing == 'unchanged':
        return False
    
    return True

def restore_std_out(outfile, saveout, scene, current_frame):
    # restore stdout
    outfile.flush()
    outfile.flush()
    outfile.close()
    sys.stdout = saveout
    if scene is not None:
        scene.frame_set(current_frame)    


def add_to_group(hierarchy_processing, custom_keywords, obj, parent_object_keyword, parent_group, groups_list, processed_objects_list, LODlevel_parent, LODdist_parent, visible_children,scene):
    # Sort objects in groups according to LOD, keyword,... : group_wrapper objects are created with parent group wrapper and a list of children group_wrapper.
    #print(" add_group name={:s} // parent_object_keyword={:s}".format(obj.name, parent_object_keyword))
    if obj.type == 'EMPTY' and len(obj.children) == 0:
        # Empty with no children: drop it.
        return
    if obj.type in {'MESH', 'EMPTY'}:
        if obj in processed_objects_list:
            return
        #print(" add_group 2")
        obj_group = None
        LODlevel, LODdist, group_name = getLOD(obj.name) # MAJ3
        keyword, keyword_type = has_reserved_name(group_name, custom_keywords) # MAJ3 / 1.4.2
        #print(" keyword 1 ", LODlevel, LODdist, keyword, keyword_type, parent_group, parent_object_keyword)
        if (keyword == None or keyword_type == KEYWORD_CUSTOM) and len(obj.children) == 0:
            # LODlevel == LODlevel_parent and LODdist != LODdist_parent and parent_object_keyword != None (example: 1_0100_Parent wwith 2 children 1_0030_Child1 and |_1_0030_Child2, then 1_0030_Child* are merged)
            if parent_object_keyword == None or keyword_type == KEYWORD_CUSTOM or (LODlevel == LODlevel_parent and LODdist != LODdist_parent and parent_object_keyword != None): # 1.4.4
                for current_group in groups_list:
                    if same_group(hierarchy_processing, current_group, LODlevel, LODdist, LODdist_parent, keyword, keyword_type, obj.parent):
                        #print(" keyword 2a ", LODlevel, LODdist, keyword, keyword_type, parent_group)
                        obj_group = current_group
                        break;
            else:
                #print(" keyword 2b ", LODlevel_parent, LODdist_parent, keyword, keyword_type, parent_group)
                if hierarchy_processing != 'unchanged':
                    if LODlevel == LODlevel_parent and LODdist == LODdist_parent:
                        obj_group = parent_group
        #print(" keyword 3 ", keyword, keyword_type, obj_group, parent_group)
        if obj_group == None:
        # Create a new group
            obj_group = group_wrapper(LODlevel, LODdist, keyword, None, group_name)
            #print(" ---- >>>> New group {:s} (keyword {:s})".format(group_name, obj_group.keyword))
            if parent_group:
                #print(" ---- >>>>--- parent_group : child_groups_list.append {:s} {:s}".format(parent_group.name, obj_group.name))
                parent_group.child_groups_list.append(obj_group) # Store child group wrapper reference in parent group wrapper
                obj_group.parent_group = parent_group # Store parent group wrapper reference in new group
            parent_group = obj_group
            groups_list.append(obj_group)
            if keyword:
                parent_object_keyword = keyword
        obj_group.objects_list.append(obj)
        #print(" >>> object={:s} obj_group={:s}".format(obj.name, obj_group.name))
        #obj_group.short_dump()
    else:
        #print(" not mesh or empty: {:s}".format(obj.name))
        parent_group = None # Example: no parent for children of LATTICE or ARMATURE
        LODlevel = '0'
        LODdist = '0000'
    processed_objects_list.append(obj)
    for obj_children in obj.children:
        #print(" child {:s} parent_object_keyword={:s} (parent {:s})".format(obj_children.name, parent_object_keyword, obj.name))
        if not visible_children or (visible_children and obj_children.is_visible(scene)): # MAJ 1.4.2
            add_to_group(hierarchy_processing, custom_keywords, obj_children, parent_object_keyword, parent_group, groups_list, processed_objects_list, LODlevel, LODdist, visible_children, scene)

        
def update_matrix(instance):
    # Update matrix to take hierarchy into account
    m, n, o = instance.mInitialTM.get_translation()
    for instance_child in instance.child_groups_list:
        if instance_child.IGinstance is not None:
            instance_child.IGinstance.mTM.update_rotation(instance.mInitialTM.get_matrix())
            instance_child.IGinstance.mTM.substract_translation(m, n, o)
            update_matrix(instance_child.IGinstance)
        
        
######################################################
# EXPORT
######################################################
            
def ExportIGS(Config):         

    # Current  model directory name
    model_directory = os.path.dirname(Config.FilePath)
    # Current model file name
    model_name = os.path.basename(Config.FilePath)
    # Name without suffix
    filename_strip = os.path.splitext(model_name)[0]
    
    myLogfile = model_directory + '\\' + filename_strip + '.log'
    outfile = open(myLogfile, 'w')
    saveout = sys.stdout
    sys.stdout = outfile

    
    default_config = configparser.ConfigParser()
    default_config_name = model_directory + '\\IGS_ExpModFile.txt'
    default_config.read(default_config_name)
    if not default_config.sections():
        print(' >>>> ERROR: File IGS_ExpModFile.txt empty or not found in model directory <<<<')
        restore_std_out(outfile, saveout, None, -1)
        print(' >>>> ERROR: File IGS_ExpModFile.txt empty or not found in model directory <<<<')
        return
    print_config(default_config, default_config_name)

    #Config.Verbose = True;    
    
    # Read Model specific file for Blender IA/IGS exporter
    specific_config = configparser.ConfigParser()
    specific_config_name = model_directory + '\\' + filename_strip + '_IGSExpModFile2.txt'
    specific_config_found = True
    if not os.path.exists(specific_config_name):
        print(' >>>> INFO: File {:s} not found <<<<'.format(specific_config_name))
        specific_config_name = model_directory + '\\' + 'IGSExpModFile2.txt'
        if not os.path.exists(specific_config_name):
            print(' >>>> INFO: File {:s} not found. <<<<\n >>>>       Objects hierarchy will not be changed. <<<<\n'.format(specific_config_name))
            specific_config_found = False
        else:
            print(' >>>> INFO: Using file {:s} <<<<'.format(specific_config_name))
    specific_config.read(specific_config_name)
    if specific_config_found:
        if not specific_config.sections():
            print(' >>>> INFO: File {:s} is empty <<<<'.format(specific_config_name))
            pass
        else:
            print_config(specific_config, specific_config_name)

    hierarchy_processing = ''
    if not specific_config_found: hierarchy_processing = 'unchanged'
    else:
        try:
            hierarchy_processing = specific_config.get('Miscellaneous', 'Hierarchy')
        except:
            pass    

    custom_keywords = []
    try:
        custom_keywords_config = specific_config.get('Miscellaneous', 'CustomKeyWords')
        custom_keywords_config = str(custom_keywords_config).replace(' ','') # remove spaces
        custom_keywords = custom_keywords_config.split(',')
    except:
        pass    
    
    # Start of export
    time1 = time.clock()

    if bpy.ops.object.mode_set.poll(): # check poll() to avoid exception
        bpy.ops.object.mode_set(mode='OBJECT') # Go into object mode to be able to select faces

    scene = Config.context.scene

    # Reset objects to start position for any animation
    current_frame = scene.frame_current
    #print(" current_frame ", current_frame)
    scene.frame_set(0)

    main_object_found = False
    if Config.use_selection:
        ob = None
#        objects_list = [ob for ob in scene.objects if ob.is_visible(scene) and ob.select]
        objects_list = [ob for ob in scene.objects if ob.select]
        if len(objects_list) != 0:
            # Pick first empty or mesh
            main_obj = objects_list[0]
            #print(' >>> main_obj ', main_obj.name, main_obj.type)
            if main_obj.type not in {'MESH'}:
                # Find a mesh or an empty in the selection
                for ob in objects_list:
                    #print(' >>> ob ', ob.name, ob.type)
                    if ob.type in {'MESH'}:
                        # Put main object at the head of the objects list
                        objects_list.remove(ob)
                        objects_list.insert(0, ob)
                        #print(" MainObject used = {:s}".format(ob.name))
                        main_obj = objects_list[0]
                        main_object_found = True
                        break
                if not main_object_found:
                    print(' >>>> No mesh found in selected objects.')
                    restore_std_out(outfile, saveout, scene, current_frame)
                    print(' >>>> No mesh found in selected objects.')
                    return
        else:
            print(' >>>> Nothing was selected')
            restore_std_out(outfile, saveout, scene, current_frame)
            print(' >>>> Nothing was selected')
            return
    else:
        # Look for main object
        main_obj = None
        main_obj_name = None
        objects_list = [ob for ob in scene.objects if ob.is_visible(scene)] # ob.is_visible(scene) => objects with eye in outliner not greyed out
        try:
            main_obj_name = specific_config.get('Miscellaneous', 'MainObject')
            print(" MainObject to find = {:s}".format(main_obj_name))
            for ob in objects_list:
                if ob.name == main_obj_name:
                    main_object_found = True
                    if ob.type in {'MESH'}:
                        # Put main object at the head of the objects list
                        objects_list.remove(ob)
                        objects_list.insert(0, ob)
                        print(" MainObject found = {:s}".format(ob.name))
                        main_obj = objects_list[0]
                        break
                    else:
                        print(" ERROR: MainObject found = {:s} but it's not a mesh: ignored".format(ob.name))
        except:
            pass
        if main_obj is None:
            if main_obj_name is None:
                print(" MainObject not specified")
            else:
                if not main_object_found:
                    print(" MainObject {:s} not found or not visible".format(main_obj_name))
            for ob in objects_list:
                #print(" Main object ??? = {:s}".format(ob.name))
                if ob.type in {'MESH'}:
                    # Take firts mesh object found
                    objects_list.remove(ob)
                    objects_list.insert(0, ob)
                    print(" Main object selected = {:s}".format(ob.name))
                    main_obj = objects_list[0]
                    break
        if main_obj is None:
            print(" No mesh or empty objects found. Export stopped")
            return

    applied_translation = Vector((0.0, 0.0, 0.0))
    applied_translation_ignored = False
    if main_object_found:
        try:
            center_main_object = specific_config.get('Miscellaneous', 'CenterMainObject')
            center_main_object = str(center_main_object).replace(' ','') # remove spaces
            s = center_main_object.split(',')
            if len(s) != 3:
                print(" ERROR: CenterMainObject ({:s}) not set with 3 floats or 'N' separated by a comma: CenterMainObject ignored".format(center_main_object))
            else:
                applied_translation = main_obj.matrix_world.to_translation()
                for num_axis in range(3):
                    if s[num_axis] == "n" or s[num_axis] == "N":
                        applied_translation[num_axis] = 0
                    else:
                        string_error = False
                        if s[num_axis][0]=='+': s[num_axis]=s[num_axis][1:]
                        if '-' in s[num_axis] and s[num_axis][0] != '-':
                            string_error = True
                        if '+' in s[num_axis]:
                            string_error = True
                        new_s = s[num_axis].replace('-','',1)
                        if not new_s.replace('.','',1).isdigit() or string_error:
                            if num_axis == 0:
                                print(" ERROR: CenterMainObject ({:s}) ignored (non numeric value found for X)".format(center_main_object))
                            if num_axis == 1:
                                print(" ERROR: CenterMainObject ({:s}) ignored (non numeric value found for Y)".format(center_main_object))
                            if num_axis == 2:
                                print(" ERROR: CenterMainObject ({:s}) ignored (non numeric value found for Z)".format(center_main_object))
                            applied_translation_ignored = True
                        else:
                            applied_translation[num_axis] -= float(s[num_axis])
            
            if applied_translation_ignored:
                applied_translation = Vector((0.0, 0.0, 0.0))
            else:
                print("\n INFO: Using CenterMainObject ({:s},{:s},{:s}), exported model moved by X={:f} Y={:f} Z={:f} (Blender axis)\n".format(s[0], s[1], s[2], \
                                                                        -applied_translation.x, -applied_translation.y, -applied_translation.z))
            
        except:
            pass    

        
    #print(" Main object selected = {:s}".format(main_obj.name))
    #print(" >> Start (sort)")
    
    # Sort objects in groups
    groups_list = []
    processed_objects_list = []
    parent_object_keyword = None
    parent_group = None
    for obj in objects_list:
        if not obj.parent: # don't process a child before its parent
            add_to_group(hierarchy_processing, custom_keywords, obj, parent_object_keyword, parent_group, groups_list, processed_objects_list, '0', '0000', Config.visible_children, scene)


    print("")
    print(" ------------------ Groups list -----------------")
    group_wrapper_dump(groups_list)
    
    '''-----------------------------------------------'''
    '''Save the sorted Blender scene into igs objects.'''
    '''-----------------------------------------------'''

    materials_list = []
    meshes_list = []
    meshesLODs_list = []
    instances_list = []
    triangles_list = []
    vertices_list = []
    textures_list = []

    current_id = 0
    triangles_count = 0
    vertices_count = 0
    meshes_count = 0
    materials_count = 0
    mExtraData = _IGDataBlockContainer(0, 0, b'', [0, 0])
    
    add_face_tag_description = False
    
    for grp in groups_list:
        #print(' >>> group ', grp.name)
        obj_matrix = Matrix() # Identity matrix
        first_obj_matrix = Matrix()
        first_obj_transform = Matrix()
        first_new_pivot = Vector([0,0,0])
        object_found = False
        Vmin_bounding_box = Vector([0,0,0])
        Vmax_bounding_box = Vector([0,0,0])
        object_triangles_count = 0
        object_triangles_list = [] # At group level = at igs mesh level
        grp_vertices_count = 0
        triangles_count_start = triangles_count # At group level
        vertices_count_start = vertices_count # At group level
        object_vertices_count_start = vertices_count # At object level
        for obj_num, obj in enumerate(grp.objects_list):
            # Process object
            object_vertices_count = 0
            object_vertices_list = [] # At object level
            object_vertices_params_list = [] # MAJ2
            #print(' >>>?? AV triangles_list len , vertices_count_start', len(triangles_list), object_vertices_count_start)
            return_code, object_vertices_count, object_triangles_count, materials_count, first_obj_matrix, first_new_pivot, first_obj_transform, object_found, Vmin_bounding_box, Vmax_bounding_box, add_face_tag_description_mesh = \
            process_mesh_geometry(bpy, grp, obj, obj_matrix, first_obj_matrix, first_new_pivot, first_obj_transform, object_found, \
            applied_translation, Vmin_bounding_box, Vmax_bounding_box, materials_list, \
            textures_list, object_triangles_list, object_vertices_list, object_vertices_count, object_vertices_count_start, \
            object_triangles_count, materials_count, default_config, specific_config, mExtraData, Config.smaller_IGS_file, object_vertices_params_list) # MAJ2
            if add_face_tag_description_mesh:
                add_face_tag_description = True
            object_vertices_count_start = object_vertices_count_start + object_vertices_count # At object level
            #print(" >>>> End process_mesh_geometry materials_count : {:d} / object_vertices_count_start {:d} / len(object_vertices_list) {:d}".format(materials_count, object_vertices_count_start, len(object_vertices_list)))
            if return_code == END_OK and object_vertices_count != 0:
                vertices_list.extend(object_vertices_list) # At object level
            elif return_code == END_ERROR:
                restore_std_out(outfile, saveout, scene, current_frame)
                print(" ERROR: check log file")
                return {'FINISHED with error'}
            grp_vertices_count = grp_vertices_count + object_vertices_count
        if object_vertices_count != 0:
            triangles_list.extend(object_triangles_list)  # At group level
            #print(' >>>?? triangles_list len , object_vertices_count', len(triangles_list), object_vertices_count)
            vertices_count = vertices_count + grp_vertices_count
    
        if object_found:
            # Create mesh related items
            new_IGfMesh, new_geomLOD, new_instance = create_mesh_igs(grp, first_obj_matrix, current_id, grp_vertices_count, object_triangles_count, \
            triangles_count_start, vertices_count_start, Vmin_bounding_box, Vmax_bounding_box, mExtraData)
            if new_instance == None:
                restore_std_out(outfile, saveout, scene, current_frame)
                print(" ERROR: check log file")
                return
            grp.IGinstance_id = int(current_id)
            current_id += 1
            if grp_vertices_count != 0: # Mesh
                meshes_count += 1
                meshes_list.append(new_IGfMesh)
                meshesLODs_list.append(new_geomLOD)
                triangles_count = triangles_count + object_triangles_count
            instances_list.append(new_instance)
            #print(' >>>?? triangles_count vertices_count', triangles_count, vertices_count)
    
    if materials_count == 0:
        print(" ERROR: No material found: Export stopped")
        restore_std_out(outfile, saveout, scene, current_frame)
        print(" ERROR: No material found: Export stopped")
        return {'FINISHED with error'}
    if vertices_count == 0:
        print(" ERROR: No vertex found: Export stopped")
        restore_std_out(outfile, saveout, scene, current_frame)
        print(" ERROR: No vertex found: Export stopped")
        return {'FINISHED with error'}
    if triangles_count == 0:
        print(" ERROR: No triangle found: Export stopped")
        restore_std_out(outfile, saveout, scene, current_frame)
        print(" ERROR: No triangle found: Export stopped")
        return {'FINISHED with error'}
    if meshes_count == 0:
        print(" ERROR: No mesh found: Export stopped")
        restore_std_out(outfile, saveout, scene, current_frame)
        print(" ERROR: No mesh found: Export stopped")
        return {'FINISHED with error'}
    
    # Update data to deal with actual offsets in the future IGS file
    datablocks_count = 1
    if add_face_tag_description:
        datablocks_count = 2

    objects_count = meshes_count
    start_materials_offset = IGS_HEADER_SIZE 
    start_vertices_offset = IGS_HEADER_SIZE + materials_count * IGS_MATERIAL_SIZE 
    start_triangles_offset = start_vertices_offset + vertices_count * IGS_VERTEX_SIZE 
    start_geom_offset = start_triangles_offset + triangles_count * IGS_TRIANGLE_SIZE 
    start_geomLOD_offset = start_geom_offset + meshes_count * IGS_GEOM_SIZE 
    start_instance_offset = start_geomLOD_offset + meshes_count * IGS_GEOMLOD_SIZE 
    start_datablocks_offset = start_instance_offset + meshes_count * IGS_INSTANCE_SIZE 

    if add_face_tag_description:
        face_tag_description_start = start_datablocks_offset + (datablocks_count * IGS_DATABLOCK_SIZE)
        absolute_texture_name_start = face_tag_description_start + IGF_MAX_FACE_TAG_DESCS * IGF_MAX_OBJECT_NAME
        start_face_tag_description_offset = datablocks_count * IGS_DATABLOCK_SIZE # Relative to start of datablocks
        start_texturenames_offset = (datablocks_count * IGS_DATABLOCK_SIZE) + (IGF_MAX_FACE_TAG_DESCS * IGF_MAX_OBJECT_NAME) # start_texturenames_offset is relative to start of datablocks
    else:
        absolute_texture_name_start = start_datablocks_offset + (datablocks_count * IGS_DATABLOCK_SIZE)
        start_texturenames_offset = datablocks_count * IGS_DATABLOCK_SIZE # start_texturenames_offset is relative to start of datablocks

    # Texture file names
    texture_names = _IGTextureNames(textures_list)

    if add_face_tag_description:
        # mName, mID, mData, mSize / UniqueID = 1
        face_tag_description_data_block = _IGDataBlock('Face tag descriptions', 1, start_face_tag_description_offset, 0)
        list_of_tags = [b'IG_TAG_NONE', b'IG_TAG_HEAD', b'IG_TAG_HEAD_STUMP', b'IG_TAG_HEAD_STUMP_REPL', b'IG_TAG_ARM_L', \
        b'IG_TAG_ARM_L_STUMP', b'IG_TAG_ARM_L_STUMP_REPL', \
        b'IG_TAG_ARM_R', b'IG_TAG_ARM_R_STUMP', b'IG_TAG_ARM_R_STUMP_REPL', b'IG_TAG_LEG_L', b'IG_TAG_LEG_L_STUMP', b'IG_TAG_LEG_L_STUMP_REPL', \
        b'IG_TAG_LEG_R', b'IG_TAG_LEG_R_STUMP', b'IG_TAG_LEG_R_STUMP_REPL']
        face_tag_description = _IGFaceTagDescription(list_of_tags)

    # mName, mID, mData, mSize / UniqueID = 2
    textures_data_block = _IGDataBlock('Texture name table', 2, start_texturenames_offset, 0)
    
    # Initialize Header structure
    object_color = _igs_rgba_color(0.4,0.4,0.4,0.0)
    blender_version = bpy.app.version_string
    exporter_version = EXPORTER_VERSION
    myComment = 'Blender ' + blender_version + ' - ' + exporter_version
    myHeader = _IGFHeader(myComment, 0, objects_count, start_instance_offset, meshes_count, start_geom_offset, 0, 0, materials_count, start_materials_offset, 0, 0, \
    object_color, 0, 0, 0, 0, datablocks_count, 0, start_datablocks_offset)

    
    # Update offset values
    # --------------------
    for triangles in triangles_list:
        #print(' Nouveau triangle')
        for i in range(3):
            #print(' >>>??? triangles.mVertex[i]', triangles.mVertex[i])
            triangles.mVertex[i] = start_vertices_offset + triangles.mVertex[i] * IGS_VERTEX_SIZE

    for mesh in meshes_list:
        # mIsSkinned, mNumVertices, mVertices, mNumTriangles, mTriangles, mBoundingBox, mExtraData
        mesh.mVertices = start_vertices_offset + mesh.mVertices * IGS_VERTEX_SIZE
        mesh.mTriangles = start_triangles_offset + mesh.mTriangles * IGS_TRIANGLE_SIZE

    for meshLOD in meshesLODs_list:
        # VisibleDistance, pGeometry, mExtraData
        meshLOD.pGeometry = start_geom_offset + meshLOD.pGeometry * IGS_GEOM_SIZE

    for instance in instances_list:
        instance.mGeometry = start_geomLOD_offset + instance.mGeometry * IGS_GEOMLOD_SIZE
        
    # Update instances for parent / children offset
    # ---------------------------------------------
    for instance in instances_list:
        # Parent offset
        if instance.parent_group is None:
            instance.mParent = 0
        else:
            instance.mParent = start_instance_offset + instance.parent_group.IGinstance_id * IGS_INSTANCE_SIZE
        # Children offset
        #print(' >>>> ', instance.mName, len(instance.child_groups_list))
        for child_num, child in enumerate(instance.child_groups_list):
            # child_groups_list is a list of 'group_wrapper' / mChildren was initialized to empty list
            #print(' >>>> ', child.name, child_num, instance.child_groups_list[child_num].IGinstance_id)
            instance.mChildren.append(start_instance_offset + instance.child_groups_list[child_num].IGinstance_id * IGS_INSTANCE_SIZE)

    # Update matrix to take hierarchy into account
    # (igs rotation and translation in a child object are relative to its parent)
    # ---------------------------------------------
    for instance in instances_list:
        if instance.parent_group is None:
        # Start from a root instance
            update_matrix(instance)
            
    '''------------------------------------------'''
    '''Dump all the igs objects to the igs file .'''
    '''------------------------------------------'''
    
    # Open the file for writing igs data
    file = open(model_directory + '\\' + filename_strip + '.igs', 'wb')
    # 
   
    myHeader.write(file)
    #if Config.Verbose:
    myHeader.dump_log(Config.Verbose)

    print('================================================================================')
    print('Materials list ({:d} materials):'.format(materials_count))
    print('--------------------------------------------------------------------------------')
    print(" {:^20} | {:^30} | {:^40} / {:s}".format("Name", "Shader", "First texture", "UVArguments"))
    for i in range(materials_count):
        materials_list[i][1].dump_log(i+1, materials_count, start_materials_offset + i * IGS_MATERIAL_SIZE, texture_names, False, Config.Verbose)
    print('================================================================================')
    if Config.Verbose:
        print('IGfMaterials                  : [ {:d} ]'.format(materials_count))
    for i in range(materials_count):
        materials_list[i][1].write(file)
        materials_list[i][1].dump_log(i+1, materials_count, start_materials_offset + i * IGS_MATERIAL_SIZE, texture_names, True, Config.Verbose)
    if Config.Verbose:
        print('================================================================================')
        print('IGfVerts                      : [ {:d} ]'.format(vertices_count))
    for i in range(vertices_count):
        vertices_list[i].write(file)
        if Config.Verbose:
            vertices_list[i].dump_log(i+1, vertices_count, start_vertices_offset + i * IGS_VERTEX_SIZE )
    if Config.Verbose:
        print('================================================================================')
        print('IGfTriangles                  : [ {:d} ]'.format(triangles_count))
    for i in range(triangles_count):
        triangles_list[i].write(file)
        if Config.Verbose:
            triangles_list[i].dump_log(i+1, triangles_count, start_triangles_offset + i * IGS_TRIANGLE_SIZE )
    if Config.Verbose:
        print('================================================================================')
        print('IGfMeshes                     : [ {:d} ]'.format(meshes_count))
    for i in range(meshes_count):
        meshes_list[i].write(file)
        if Config.Verbose:
            meshes_list[i].dump_log(i+1, meshes_count, start_geom_offset + i * IGS_GEOM_SIZE)
    if Config.Verbose:
        print('================================================================================')
        print('IGfMeshLODs                   : [ {:d} ]'.format(meshes_count))
    for i in range(meshes_count):
        meshesLODs_list[i].write(file)
        if Config.Verbose:
            meshesLODs_list[i].dump_log(i+1, meshes_count, start_geomLOD_offset + i * IGS_GEOMLOD_SIZE)
    if Config.Verbose:
        print('================================================================================')
        print('IGfObjects                    : [ {:d} ]'.format(meshes_count))
    for i in range(meshes_count):
        instances_list[i].write(file)
        if Config.Verbose:
            instances_list[i].dump_log(i+1, meshes_count, start_instance_offset + i * IGS_INSTANCE_SIZE)
    if Config.Verbose:
        print('================================================================================')
        print('IGfDataBlocks                 : [ {:d} ]'.format(datablocks_count))
        num_block = 1
        if add_face_tag_description:
            face_tag_description_data_block.dump_log(1, datablocks_count, start_datablocks_offset)
            num_block = 2
        textures_data_block.dump_log(num_block, datablocks_count, start_datablocks_offset + (num_block - 1) * IGS_DATABLOCK_SIZE)
    if add_face_tag_description:
        face_tag_description_data_block.write(file)
    textures_data_block.write(file)
    if add_face_tag_description:
        if Config.Verbose:
            face_tag_description.dump_log(start_datablocks_offset + 2 * IGS_DATABLOCK_SIZE)
        face_tag_description.write(file)
    texture_names.write(file)

    print('================================================================================')
    texture_names.dump_log(1, 1, start_texturenames_offset)

    # Close the file:
    file.close()
    
    print("\n\nIGS export time: {:.02f}".format(time.clock() - time1))

#---------------------------
    restore_std_out(outfile, saveout, scene, current_frame)

    # open the log file for output to console
    '''    try:
        infile = open(myLogfile, 'r')
        for line in infile:
            print(line, end="")
    except:
        print('Error reading log file')
    else:
        infile.close()'''
 #---------------------------

    return {'FINISHED'}


class IGSExporter(bpy.types.Operator):
    """Export to the IGS model format (.igs) for Railworks"""

    bl_idname = "export.igs"
    bl_label = "Export IGS"

    filepath = StringProperty(subtype='FILE_PATH')

    # Options
    use_selection = BoolProperty(
        name="Use selection",
        description="Enable to export only the selected objects. Only Mesh and Empty " \
                    "objects will be processed for export. Children of the selected objects " \
                    "are exported according to the \"visible children\" option.",
        default=False)

    visible_children = BoolProperty(
        name="Visible children",
        description="Enable to export only the visible children.",
        default=False)

    smaller_IGS_file = BoolProperty(
        name="Optimized IGS",
        description="Enable to export a smaller IGS file by storing only once points shared by different faces of the same object. " \
                    "It is compulsory for some shaders.",
        default=True)

    Verbose = BoolProperty(
        name="Verbose",
        description="Detailed log file. Check the console for output",
        default=False)

    def execute(self, context):
        # Append .igs
        FilePath = bpy.path.ensure_ext(self.filepath, ".igs")

        Config = IGSExporterSettings(context,
                                     FilePath,
                                     smaller_IGS_file=self.smaller_IGS_file,
                                     use_selection=self.use_selection,
                                     visible_children=self.visible_children,
                                     Verbose=self.Verbose)

        ExportIGS(Config)
        return {"FINISHED"}

    def invoke(self, context, event):
        if not self.filepath:
            self.filepath = bpy.path.ensure_ext(bpy.data.filepath, ".igs")
        WindowManager = context.window_manager
        WindowManager.fileselect_add(self)
        return {"RUNNING_MODAL"}


def menu_func(self, context):
    self.layout.operator(IGSExporter.bl_idname, text="TS20xx IGS (.igs)")


def register():
    bpy.utils.register_module(__name__)

    bpy.types.INFO_MT_file_export.append(menu_func)


def unregister():
    bpy.utils.unregister_module(__name__)

    bpy.types.INFO_MT_file_export.remove(menu_func)


if __name__ == "__main__":
    register()
    